# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Here I collect all functions needed to parse the output of a KKR calculation.
These functions do not need aiida and are therefore separated from the actual
parser file where parse_kkr_outputfile is called
"""
from numpy import ndarray, array, loadtxt, shape
from masci_tools.io.common_functions import (search_string, get_version_info, angles_to_vec,
                                             get_corestates_from_potential, get_highest_core_state, open_general,
                                             convert_to_pystd)
from masci_tools.util.constants import RY_TO_EV
import traceback

__copyright__ = (u'Copyright (c), 2017, Forschungszentrum Jülich GmbH,' 'IAS-1/PGI-1, Germany. All rights reserved.')
__license__ = 'MIT license, see LICENSE.txt file'
__contributors__ = u'Philipp Rüßmann'
__version__ = '1.8'

####################################################################################


def parse_array_float(outfile, searchstring, splitinfo, replacepair=None, debug=False):
    """
    Search for keyword `searchstring` in `outfile` and extract array of results

    .. note:
        `splitinfo` can be of the form [1, 'X', 1] or [2, 'X', 1, 0] where
        splitinfo[0] can only be 1 or 2 (determines the mode),
        splitinfo[1] is the string at which the line is split,
        splitinfo[2] is the index which is used,
        splitinfo[3] (only for splitinfo[0]==2) is the part that is taken after applying split() a second time (split at whitespace).

    .. note:
        If `replacepair` is not None the substring replacepair[0] is replaced by replacepair[1] before processing further

    Returns: array of results

    """
    f = open_general(outfile)
    tmptxt = f.readlines()
    f.close()
    itmp = 0
    res = []
    while itmp >= 0:
        itmp = search_string(searchstring, tmptxt)
        if debug:
            print(('in parse_array_float (itmp, searchstring, outfile):', itmp, searchstring, outfile))
        if itmp >= 0:
            tmpval = tmptxt.pop(itmp)
            if replacepair is not None:
                tmpval = tmpval.replace(replacepair[0], replacepair[1])
            if splitinfo[0] == 1:
                tmpval = float(tmpval.split(splitinfo[1])[splitinfo[2]])
            elif splitinfo[0] == 2:
                tmpval = float(tmpval.split(splitinfo[1])[splitinfo[2]].split()[splitinfo[3]])
            else:
                raise ValueError('splitinfo[0] has to be either 1 or 2')
            res.append(tmpval)
    res = array(res)
    return res


def get_rms(outfile, outfile2, debug=False):
    """
    Get rms error per atom (both values for charge and spin) and total (i.e. average) value
    """
    if debug:
        print((outfile, outfile2))
    rms_charge = parse_array_float(outfile, 'average rms-error', [2, '=', 1, 0], ['D', 'E'], debug=debug)
    if debug:
        print(rms_charge)
    rms_spin = parse_array_float(
        outfile, 'v+ - v-', [1, '=', 1],
        ['D', 'E'])  # this should be in the line after 'average rms-error' but is only present if NSPIN==2
    if debug:
        print(rms_spin)
    rms_charge_atoms = parse_array_float(outfile2, 'rms-error for atom', [2, '=', 1, 0], ['D', 'E'])
    if debug:
        print(rms_charge_atoms)
    rms_spin_atoms = parse_array_float(outfile2, 'rms-error for atom', [2, '=', 1, 0],
                                       ['D', 'E'])  # only present for NSPIN==2
    if debug:
        print(rms_spin_atoms)
    niter = len(rms_charge)  # number of iterations
    if debug:
        print(niter)
    natoms = int(len(rms_charge_atoms) //
                 niter)  # number of atoms in system, needed to take only atom resolved rms of last iteration
    if debug:
        print(natoms)
    return rms_charge, rms_spin, rms_charge_atoms[-natoms:], rms_spin_atoms[-natoms:]


def get_noco_rms(outfile, debug=False):
    """
    Get average noco rms error
    """
    if debug:
        print(outfile)
    try:
        rms_noco = parse_array_float(outfile, 'Total RMS(angles)', [1, ':', 1], debug=debug)
    except:
        rms_noco = []
        if debug:
            traceback.print_exc()
    return rms_noco


def get_neutr(outfile):
    res = parse_array_float(outfile, 'charge neutrality in unit cell', [1, '=', 1])
    return res


def get_magtot(outfile):
    res = parse_array_float(outfile, 'TOTAL mag. moment in unit cell', [1, '=', 1])
    return res


def get_EF(outfile):
    res = parse_array_float(outfile, 'E FERMI', [2, 'FERMI', 1, 0])
    return res


def get_DOS_EF(outfile):
    res = parse_array_float(outfile, 'DOS(E_F)', [1, '=', 1])
    return res


def get_Etot(outfile):
    res = parse_array_float(outfile, 'TOTAL ENERGY in ryd.', [1, ':', 1])
    return res


def find_warnings(outfile):
    f = open_general(outfile)
    tmptxt = f.readlines()
    tmptxt_caps = [txt.upper() for txt in tmptxt]
    f.close()
    itmp = 0
    res = []
    while itmp >= 0:
        itmp = search_string('WARNING', tmptxt_caps)
        if itmp >= 0:
            tmpval = tmptxt_caps.pop(itmp)
            tmpval = tmptxt.pop(itmp)
            res.append(tmpval.strip())
    return array(res)


def extract_timings(outfile):
    f = open_general(outfile)
    tmptxt = f.readlines()
    f.close()
    itmp = 0
    res = []
    search_keys = [
        'main0',
        'main1a - tbref',
        'main1a  ',  # two spaces to differentiate from following key
        'main1b - calctref13',
        'main1b  ',  # two spaces!
        'main1c - serial part',
        'main1c  ',  # two spaces!
        'main2',
        'Time in Iteration'
    ]
    while itmp >= 0:
        tmpvals = []
        for isearch in search_keys:
            itmp = search_string(isearch, tmptxt)
            if itmp >= 0:
                tmpval = [isearch, float(tmptxt.pop(itmp).split()[-1])]
                tmpvals.append(tmpval)
        if len(tmpvals) > 0:
            res.append(tmpvals)
    res = res[0]
    return dict(res)


def get_charges_per_atom(outfile_000):
    res1 = parse_array_float(outfile_000, 'charge in wigner seitz', [1, '=', 1])
    # these two are not in output of DOS calculation (and are then ignored)
    res2 = parse_array_float(outfile_000, 'nuclear charge', [2, 'nuclear charge', 1, 0])
    try:
        res3 = parse_array_float(outfile_000, 'core charge', [1, '=', 1])
    except IndexError:
        res3 = parse_array_float(outfile_000, 'core charge', [1, ':', 1])
    return res1, res2, res3


def get_single_particle_energies(outfile_000):
    """
    extracts single particle energies from outfile_000 (output.000.txt)
    returns the valence contribution of the single particle energies
    """
    f = open_general(outfile_000)
    tmptxt = f.readlines()
    f.close()
    itmp = 0
    res = []
    while itmp >= 0:
        itmp = search_string('band energy per atom', tmptxt)
        if itmp >= 0:
            tmpval = float(tmptxt.pop(itmp).split()[-1])
            res.append(tmpval)
    return array(res)


def get_econt_info(outfile_0init):
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()

    itmp = search_string('E min', tmptxt)
    emin = float(tmptxt[itmp].split('min')[1].split('=')[1].split()[0])

    itmp = search_string('Temperature', tmptxt)
    tempr = float(tmptxt[itmp].split('Temperature')[1].split('=')[1].split()[0])

    itmp = search_string('Number of energy points', tmptxt)
    Nepts = int(tmptxt[itmp].split(':')[1].split()[0])

    doscalc = search_string('Density-of-States calculation', tmptxt)
    if doscalc == -1:
        # npol
        itmp = search_string('poles =', tmptxt)
        Npol = int(tmptxt[itmp].split('=')[1].split()[0])
        # npt1, npt2, npt3
        itmp = search_string('contour:', tmptxt)
        tmp = tmptxt[itmp].replace(',', '').replace('=', '= ').split(':')[1].split()
        N1 = int(tmp[2])
        N2 = int(tmp[5])
        N3 = int(tmp[8])
    else:
        Npol, N1, N2, N3 = 0, 0, Nepts, 0

    return emin, tempr, Nepts, Npol, N1, N2, N3


def get_core_states(potfile):
    ncore, energies, lmoments = get_corestates_from_potential(potfile=potfile)
    emax, lmax, descr_max = [], [], []
    for ipot in range(len(ncore)):
        if ncore[ipot] > 0:
            lvalmax, energy_max, descr = get_highest_core_state(ncore[ipot], energies[ipot], lmoments[ipot])
        else:
            lvalmax, energy_max, descr = None, None, 'no core states'
        emax.append(energy_max)
        lmax.append(lvalmax)
        descr_max.append(descr)
    return array(ncore), array(emax), array(lmax), array(descr_max)


def get_alatinfo(outfile_0init):
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    itmp = search_string('Lattice constants :', tmptxt)
    alat = float(tmptxt[itmp].split(':')[1].split('=')[1].split()[0])
    twopialat = float(tmptxt[itmp].split(':')[1].split('=')[2].split()[0])
    return alat, twopialat


def get_scfinfo(outfile_0init, outfile_000, outfile):
    f = open_general(outfile_000)
    tmptxt = f.readlines()
    f.close()

    itmp = search_string('ITERATION :', tmptxt)
    tmpval = tmptxt[itmp].split(':')[1].split()
    niter = int(tmpval[0])
    nitermax = int(tmpval[3])

    f = open_general(outfile)
    tmptxt = f.readlines()
    f.close()
    itmp1 = search_string('SCF ITERATION CONVERGED', tmptxt)
    itmp2 = search_string('NUMBER OF SCF STEPS EXHAUSTED', tmptxt)
    if itmp1 >= 0:
        converged = True
    else:
        converged = False
    if itmp2 >= 0:
        nmax_reached = True
    else:
        nmax_reached = False

    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    itmp = search_string('STRMIX        FCM       QBOUND', tmptxt)
    tmpval = tmptxt[itmp + 1].split()
    strmix = float(tmpval[0])
    fcm = float(tmpval[1])
    qbound = float(tmpval[2])
    tmpval = tmptxt[itmp + 4].split()
    brymix = float(tmpval[0])
    itmp = search_string('IMIX    IGF    ICC', tmptxt)
    imix = int(tmptxt[itmp + 1].split()[0])
    idtbry = int(tmptxt[itmp + 4].split()[0])

    mixinfo = [imix, strmix, qbound, fcm, idtbry, brymix]

    return niter, nitermax, converged, nmax_reached, mixinfo


def get_kmeshinfo(outfile_0init, outfile_000):
    """
    Extract kmesh info from output.0.txt and output.000.txt
    """
    # first get info from output.0.txt
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    nkmesh = []
    itmp = search_string('number of different k-meshes', tmptxt)
    nkmesh.append(int(tmptxt[itmp].split(':')[1].split()[0]))
    itmp = search_string('NofKs', tmptxt)
    nofks, nkx, nky, nkz = [], [], [], []
    if itmp >= 0:
        for ik in range(nkmesh[0]):
            tmpval = tmptxt[itmp + 2 + ik].split()
            nofks.append(int(tmpval[1]))
            nkx.append(int(tmpval[2]))
            nky.append(int(tmpval[3]))
            nkz.append(int(tmpval[4]))

    tmpdict = {'number_of_kpts': nofks, 'n_kx': nkx, 'n_ky': nky, 'n_kz': nkz}
    nkmesh.append(tmpdict)

    #next get kmesh_ie from output.000.txt
    f = open_general(outfile_000)
    tmptxt = f.readlines()
    f.close()
    kmesh_ie = []
    itmp = 0
    while itmp >= 0:
        itmp = search_string('KMESH =', tmptxt)
        if itmp >= 0:
            tmpval = int(tmptxt.pop(itmp).split()[-1])
            kmesh_ie.append(tmpval)

    return nkmesh, kmesh_ie


def get_symmetries(outfile_0init):
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    try:
        itmp = search_string('symmetries found for this lattice:', tmptxt)
        nsym = int(tmptxt[itmp].split(':')[1].split()[0])
    except IndexError:
        itmp = search_string('< FINDGROUP > : Finding symmetry operations', tmptxt)
        tmptxt2 = tmptxt[itmp:]
        itmp = search_string('found for this lattice:', tmptxt2)
        nsym = int(tmptxt2[itmp].split(':')[1].split()[0])
    itmp = search_string('symmetries will be used', tmptxt)
    nsym_used = int(tmptxt[itmp].split()[3])
    itmp = search_string('<SYMTAUMAT>', tmptxt)
    tmpdict = {}
    for isym in range(nsym_used):
        tmpval = tmptxt[itmp + 5 + isym].replace('0-',
                                                 '0 -').replace('1-',
                                                                '1 -').split()  # bugfix for -120 degree euler angle
        desc = tmpval[1]
        inversion = int(tmpval[2])
        euler = [float(tmpval[3]), float(tmpval[4]), float(tmpval[5])]
        unitary = int(tmpval[6].replace('T', '1').replace('F', '0'))
        tmpdict[desc] = {'has_inversion': inversion, 'is_unitary': unitary, 'euler_angles': euler}
    desc = tmpdict
    return nsym, nsym_used, desc


def get_ewald(outfile_0init):
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    itmp = search_string('setting bulk Madelung coefficients', tmptxt)
    if itmp >= 0:
        info = '3D'
    else:
        info = '2D'
    if info == '3D':
        itmp = search_string('< LATTICE3D >', tmptxt)
        tmpval = tmptxt[itmp + 7].split()[2:]
        rsum = float(tmpval[2]), int(tmpval[0]), int(tmpval[1])
        tmpval = tmptxt[itmp + 8].split()[2:]
        gsum = float(tmpval[2]), int(tmpval[0]), int(tmpval[1])
    else:
        itmp = search_string('< LATTICE2D >', tmptxt)
        tmpval = tmptxt[itmp + 13].split()[2:]
        rsum = float(tmpval[2]), int(tmpval[0]), int(tmpval[1])
        tmpval = tmptxt[itmp + 14].split()[2:]
        gsum = float(tmpval[2]), int(tmpval[0]), int(tmpval[1])
    return rsum, gsum, info


def get_nspin(outfile_0init):
    """
    extract NSPIN value from output.0.txt
    """
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    itmp = search_string('NSPIN', tmptxt)
    nspin = int(tmptxt[itmp + 1].split()[0])
    return nspin


def get_natom(outfile_0init):
    """
    extract NATYP value from output.0.txt
    """
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    itmp = search_string('NATYP', tmptxt)
    natom = int(tmptxt[itmp + 1].split()[0])
    return natom


def use_newsosol(outfile_0init):
    """
    extract NEWSOSOL info from output.0.txt
    """
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    newsosol = False
    # old style (RUNOPT output)
    itmp = search_string('NEWSOSOL', tmptxt)
    if itmp >= 0 and 'NEWSOSOL' in tmptxt[itmp].upper():
        newsosol = True
    itmp = search_string('<use_Chebychev_solver>=', tmptxt)
    # new style: check for output of runoptions
    if itmp >= 0:
        if tmptxt[itmp].split()[1][:1].upper() == 'T':
            newsosol = True
        if tmptxt[itmp].split()[1][:1].upper() == 'F':
            newsosol = False
    return newsosol


def get_spinmom_per_atom(outfile, natom, nonco_out_file=None):
    """
    Extract spin moment information from outfile and nonco_angles_out (if given)
    """
    f = open_general(outfile)
    tmptxt = f.readlines()
    f.close()
    itmp = 0
    result = []
    while itmp >= 0:
        itmp = search_string('m_spin', tmptxt)
        if itmp >= 0:
            tmpline = tmptxt.pop(itmp)
            tmparray = []
            for iatom in range(natom):
                tmpline = tmptxt.pop(itmp)
                tmparray.append(float(tmpline.split()[3]))
            result.append(tmparray)

    # if the file is there, i.e. NEWSOSOL is used, then extract also direction of spins (angles theta and phi)
    if nonco_out_file is not None and result != []:
        angles = loadtxt(nonco_out_file, usecols=[0, 1])  # make sure only theta and phi are read in
        if len(shape(angles)) == 1:
            angles = array([angles])
        vec = angles_to_vec(result[-1], angles[:, 0], angles[:, 1])
    else:
        vec, angles = [], []

    return array(result), vec, angles


def get_orbmom(outfile, natom):
    """
    read orbmom info from outfile and return array (iteration, atom)=orbmom
    """
    f = open_general(outfile)
    tmptxt = f.readlines()
    f.close()
    itmp = 0
    result = []
    while itmp >= 0:
        itmp = search_string('m_spin', tmptxt)
        if itmp >= 0:
            tmpline = tmptxt.pop(itmp)
            tmparray = []
            for iatom in range(natom):
                tmpline = tmptxt.pop(itmp)
                tmparray.append(float(tmpline.split()[4]))
            result.append(tmparray)

    return array(result)  #, vec, angles


def get_lattice_vectors(outfile_0init):
    """
    read direct and reciprocal lattice vectors in internal units (useful for qdos generation)
    """
    f = open_general(outfile_0init)
    tmptxt = f.readlines()
    f.close()
    vecs, rvecs = [], []
    tmpvecs = []
    for search_txt in ['a_1: ', 'a_2: ', 'a_3: ', 'b_1: ', 'b_2: ', 'b_3: ']:
        itmp = search_string(search_txt, tmptxt)
        if itmp >= 0:
            tmpvec = tmptxt[itmp].split(':')[1].split()
            tmpvecs.append([float(tmpvec[0]), float(tmpvec[1]), float(tmpvec[1])])
        if search_txt in ['a_3: ', 'b_3: '] and itmp < 0:
            # reset vecs for 2D case
            tmpvecs[0] = tmpvecs[0][:2]
            tmpvecs[1] = tmpvecs[1][:2]
        if search_txt == 'a_3: ':
            vecs = tmpvecs
            tmpvecs = []
        elif search_txt == 'b_3: ':
            rvecs = tmpvecs
    return vecs, rvecs


def parse_kkr_outputfile(out_dict,
                         outfile,
                         outfile_0init,
                         outfile_000,
                         timing_file,
                         potfile_out,
                         nonco_out_file,
                         outfile_2='output.2.txt',
                         skip_readin=False,
                         debug=False):
    """
    Parser method for the kkr outfile. It returns a dictionary with results
    """
    # scaling factors etc. defined globally
    doscalc = False

    # collection of parsing error messages
    msg_list = []

    try:
        code_version, compile_options, serial_number = get_version_info(outfile)
        tmp_dict = {}
        tmp_dict['code_version'] = code_version
        tmp_dict['compile_options'] = compile_options
        tmp_dict['calculation_serial_number'] = serial_number
        out_dict['code_info_group'] = tmp_dict
    except:
        msg = 'Error parsing output of KKR: Version Info'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        nspin = get_nspin(outfile_0init)
        natom = get_natom(outfile_0init)
        newsosol = use_newsosol(outfile_0init)
        out_dict['nspin'] = nspin
        out_dict['number_of_atoms_in_unit_cell'] = natom
        out_dict['use_newsosol'] = newsosol
    except:
        msg = 'Error parsing output of KKR: nspin/natom'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        result = find_warnings(outfile)
        tmp_dict = {}
        tmp_dict['number_of_warnings'] = len(result)
        tmp_dict['warnings_list'] = result
        out_dict['warnings_group'] = tmp_dict
    except:
        msg = 'Error parsing output of KKR: search for warnings'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        result = extract_timings(timing_file)
        out_dict['timings_group'] = result
        out_dict['timings_unit'] = 'seconds'
    except:
        msg = 'Error parsing output of KKR: timings'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        emin, tempr, Nepts, Npol, N1, N2, N3 = get_econt_info(outfile_0init)
        tmp_dict = {}
        tmp_dict['emin'] = emin
        tmp_dict['emin_unit'] = 'Rydberg'
        tmp_dict['number_of_energy_points'] = Nepts
        tmp_dict['temperature'] = tempr
        tmp_dict['temperature_unit'] = 'Kelvin'
        tmp_dict['npol'] = Npol
        tmp_dict['n1'] = N1
        tmp_dict['n2'] = N2
        tmp_dict['n3'] = N3
        out_dict['energy_contour_group'] = tmp_dict
        if Npol == 0:
            doscalc = True
    except:
        msg = 'Error parsing output of KKR: energy contour'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        alat, twopioveralat = get_alatinfo(outfile_0init)
        out_dict['alat_internal'] = alat
        out_dict['two_pi_over_alat_internal'] = twopioveralat
        out_dict['alat_internal_unit'] = 'a_Bohr'
        out_dict['two_pi_over_alat_internal_unit'] = '1/a_Bohr'
    except:
        msg = 'Error parsing output of KKR: alat, 2*pi/alat'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        nkmesh, kmesh_ie = get_kmeshinfo(outfile_0init, outfile_000)
        tmp_dict = {}
        tmp_dict['number_different_kmeshes'] = nkmesh[0]
        tmp_dict['number_kpoints_per_kmesh'] = nkmesh[1]
        tmp_dict['kmesh_energypoint'] = kmesh_ie
        out_dict['kmesh_group'] = tmp_dict
    except:
        msg = 'Error parsing output of KKR: kmesh'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    try:
        nsym, nsym_used, desc = get_symmetries(outfile_0init)
        tmp_dict = {}
        tmp_dict['number_of_lattice_symmetries'] = nsym
        tmp_dict['number_of_used_symmetries'] = nsym_used
        tmp_dict['symmetry_description'] = desc
        out_dict['symmetries_group'] = tmp_dict
    except:
        msg = 'Error parsing output of KKR: symmetries'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    if not doscalc:  # in case of dos calculation no ewald summation is done
        try:
            rsum, gsum, info = get_ewald(outfile_0init)
            tmp_dict = {}
            tmp_dict['ewald_summation_mode'] = info
            tmp_dict['rsum_cutoff'] = rsum[0]
            tmp_dict['rsum_number_of_vectors'] = rsum[1]
            tmp_dict['rsum_number_of_shells'] = rsum[2]
            tmp_dict['rsum_cutoff_unit'] = 'a_Bohr'
            tmp_dict['gsum_cutoff'] = gsum[0]
            tmp_dict['gsum_number_of_vectors'] = gsum[1]
            tmp_dict['gsum_number_of_shells'] = gsum[2]
            tmp_dict['gsum_cutoff_unit'] = '1/a_Bohr'
            out_dict['ewald_sum_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: ewald summation for madelung poterntial'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

    try:
        bv, recbv = get_lattice_vectors(outfile_0init)
        out_dict['direct_bravais_matrix'] = bv
        out_dict['reciprocal_bravais_matrix'] = recbv
        out_dict['direct_bravais_matrix_unit'] = 'alat'
        out_dict['reciprocal_bravais_matrix_unit'] = '2*pi / alat'
    except:
        msg = 'Error parsing output of KKR: lattice vectors (direct/reciprocal)'
        msg_list.append(msg)
        if debug:
            traceback.print_exc()

    # this is skipped for qdos run for example
    if not skip_readin:
        try:
            ncore, emax, lmax, descr_max = get_core_states(potfile_out)
            tmp_dict = {}
            tmp_dict['number_of_core_states_per_atom'] = ncore
            tmp_dict['energy_highest_lying_core_state_per_atom'] = emax
            tmp_dict['energy_highest_lying_core_state_per_atom_unit'] = 'Rydberg'
            tmp_dict['descr_highest_lying_core_state_per_atom'] = descr_max
            out_dict['core_states_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: core_states'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        tmp_dict = {}  # used to group convergence info (rms, rms per atom, charge neutrality)
        # also initialize convegence_group where all info stored for all iterations is kept
        out_dict['convergence_group'] = tmp_dict
        try:
            rms_charge, rms_spin, result_atoms_last_charge, result_atoms_last_spin = get_rms(outfile,
                                                                                             outfile_000,
                                                                                             debug=debug)
            tmp_dict['rms'] = rms_charge[-1]
            tmp_dict['rms_all_iterations'] = rms_charge
            tmp_dict['rms_per_atom'] = result_atoms_last_charge
            if len(rms_spin) > 0:
                tmp_dict['rms_spin'] = rms_spin[-1]
            else:
                tmp_dict['rms_spin'] = None
            tmp_dict['rms_spin_all_iterations'] = rms_spin
            tmp_dict['rms_spin_per_atom'] = result_atoms_last_spin
            tmp_dict['rms_unit'] = 'unitless'
            out_dict['convergence_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: rms-error'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        try:
            result = get_neutr(outfile)
            tmp_dict['charge_neutrality'] = result[-1]
            out_dict['convergence_group']['charge_neutrality_all_iterations'] = result
            tmp_dict['charge_neutrality_unit'] = 'electrons'
            out_dict['convergence_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: charge neutrality'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        tmp_dict = {}  # used to group magnetism info (spin and orbital moments)
        try:
            result = get_magtot(outfile)
            if len(result) > 0:
                tmp_dict['total_spin_moment'] = result[-1]
                out_dict['convergence_group']['total_spin_moment_all_iterations'] = result
                tmp_dict['total_spin_moment_unit'] = 'mu_Bohr'
                out_dict['magnetism_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: total magnetic moment'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        try:
            if nspin > 1:
                if not newsosol:
                    #reset automatically to None to turn off reading of nonco angles file
                    nonco_out_file = None

                result, vec, angles = get_spinmom_per_atom(outfile, natom, nonco_out_file)
                if len(result) > 0:
                    tmp_dict['spin_moment_per_atom'] = result[-1, :]
                    if newsosol:
                        tmp_dict['spin_moment_vector_per_atom'] = vec[:]
                        tmp_dict['spin_moment_angles_per_atom'] = angles[:]
                        tmp_dict['spin_moment_angles_per_atom_unit'] = 'degree'
                    out_dict['convergence_group']['spin_moment_per_atom_all_iterations'] = result[:, :]
                    tmp_dict['spin_moment_unit'] = 'mu_Bohr'
                    out_dict['magnetism_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: spin moment per atom'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        # add orbital moments to magnetis group in parser output
        try:
            if nspin > 1 and newsosol:
                #TODO orbital moment full vectors
                # so far the KKR code writes only the component of the orbital moment
                # parallel to the spin moment, thus vec and angles are returned empty
                # by construction. This might change in the future
                #result, vec, angles = get_orbmom(outfile, natom, nonco_angles_orbmom)
                # so for now return only result= array containing all iterations, all atoms, orbital moment parallel to spin quantization axis
                result = get_orbmom(outfile, natom)
                if len(result) > 0:
                    tmp_dict['total_orbital_moment'] = sum(result[-1, :])
                    tmp_dict['orbital_moment_per_atom'] = result[-1, :]
                    #tmp_dict['orbital_moment_vector_per_atom'] = vec[-1,:]
                    #tmp_dict['orbital_moment_angles_per_atom'] = angles[-1,:]
                    out_dict['convergence_group']['orbital_moment_per_atom_all_iterations'] = result[:, :]
                    tmp_dict['orbital_moment_unit'] = 'mu_Bohr'
                    #tmp_dict['orbital_moment_angles_per_atom_unit'] = 'degree'
                    out_dict['magnetism_group'] = tmp_dict
        except:
            msg = 'Error parsing output of KKR: orbital moment'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        # get RMS info for nonco angles
        try:
            if nspin > 1 and newsosol:
                result = get_noco_rms(outfile, debug)
                if len(result) > 0:
                    out_dict['convergence_group']['noco_angles_rms_all_iterations'] = result[:]
                    out_dict['convergence_group']['noco_angles_rms_all_iterations_unit'] = 'degrees'
        except:
            msg = 'Error parsing output of KKR: noco angles rms value'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        try:
            result = get_EF(outfile)
            out_dict['fermi_energy'] = result[-1]
            out_dict['fermi_energy_units'] = 'Ry'
            out_dict['convergence_group']['fermi_energy_all_iterations'] = result
            out_dict['convergence_group']['fermi_energy_all_iterations_units'] = 'Ry'
        except:
            msg = 'Error parsing output of KKR: EF'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        try:
            result = get_DOS_EF(outfile)
            out_dict['dos_at_fermi_energy'] = result[-1]
            out_dict['convergence_group']['dos_at_fermi_energy_all_iterations'] = result
        except:
            msg = 'Error parsing output of KKR: DOS@EF'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

        try:
            result = get_Etot(outfile)
            out_dict['energy'] = result[-1] * RY_TO_EV
            out_dict['energy_unit'] = 'eV'
            out_dict['total_energy_Ry'] = result[-1]
            out_dict['total_energy_Ry_unit'] = 'Rydberg'
            out_dict['convergence_group']['total_energy_Ry_all_iterations'] = result
        except:
            msg = 'Error parsing output of KKR: total energy'
            msg_list.append(msg)
        if debug:
            traceback.print_exc()

        try:
            result = get_single_particle_energies(outfile_000)
            out_dict['single_particle_energies'] = result * RY_TO_EV
            out_dict['single_particle_energies_unit'] = 'eV'
        except:
            if not doscalc:
                msg = 'Error parsing output of KKR: single particle energies'
                msg_list.append(msg)
                if debug:
                    traceback.print_exc()

        try:
            result_WS, result_tot, result_C = get_charges_per_atom(outfile_000)
            niter = len(out_dict['convergence_group']['rms_all_iterations'])
            natyp = int(len(result_tot) // niter)
            out_dict['total_charge_per_atom'] = result_tot[-natyp:]
            out_dict['charge_core_states_per_atom'] = result_C[-natyp:]
            # this check deals with the DOS case where output is slightly different
            if len(result_WS) == len(result_C):
                out_dict['charge_valence_states_per_atom'] = result_WS[-natyp:] - result_C[-natyp:]
            out_dict['total_charge_per_atom_unit'] = 'electron charge'
            out_dict['charge_core_states_per_atom_unit'] = 'electron charge'
            out_dict['charge_valence_states_per_atom_unit'] = 'electron charge'
        except:
            if not doscalc:
                msg = 'Error parsing output of KKR: charges'
                msg_list.append(msg)
                if debug:
                    traceback.print_exc()

        try:
            try:
                niter, nitermax, converged, nmax_reached, mixinfo = get_scfinfo(outfile_0init, outfile_000, outfile)
            except IndexError:
                niter, nitermax, converged, nmax_reached, mixinfo = get_scfinfo(outfile_0init, outfile_2, outfile)
            out_dict['convergence_group']['number_of_iterations'] = niter
            out_dict['convergence_group']['number_of_iterations_max'] = nitermax
            out_dict['convergence_group']['calculation_converged'] = converged
            out_dict['convergence_group']['nsteps_exhausted'] = nmax_reached
            out_dict['convergence_group']['imix'] = mixinfo[0]
            out_dict['convergence_group']['strmix'] = mixinfo[1]
            out_dict['convergence_group']['qbound'] = mixinfo[2]
            out_dict['convergence_group']['fcm'] = mixinfo[3]
            out_dict['convergence_group']['idtbry'] = mixinfo[4]
            out_dict['convergence_group']['brymix'] = mixinfo[5]
        except:
            msg = 'Error parsing output of KKR: scfinfo'
            msg_list.append(msg)
            if debug:
                traceback.print_exc()

    #convert numpy arrays to standard python lists
    out_dict = convert_to_pystd(out_dict)

    # return output with error messages if there are any
    return len(msg_list) == 0, msg_list, out_dict


def check_error_category(err_cat, err_msg, out_dict):
    """
    Check if parser error of the non-critical category (err_cat != 1) are
    actually consistent and may be discarded.

    :param err_cat: the error-category of the error message to be investigated
    :param err_msg: the error-message
    :param out_dict: the dict of results obtained from the parser function

    :returns: True/False if message is an error or warning
    """
    # check special cases:
    # 1. nonco_angle_file not present, but newsosol==False anyways
    if 'NONCO_ANGLES_OUT' in err_msg:
        if 'use_newsosol' in out_dict:
            return out_dict['use_newsosol']
        else:
            return True

    # default behavior
    return err_cat == 1



# -*- coding: utf-8 -*-
"""
Tests of the calculator of mathematical expressions in the inp.xml files
"""
import pytest
from masci_tools.util.fleur_calculate_expression import calculate_expression
from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
import numpy as np


def test_calculate_expression():

    assert pytest.approx(calculate_expression('1.0+3.0-2.45*8.45', FLEUR_DEFINED_CONSTANTS)) == -16.7025
    assert pytest.approx(calculate_expression('Pi/4.0', FLEUR_DEFINED_CONSTANTS)) == 0.78539816339
    assert pytest.approx(calculate_expression('cos(Pi)-exp(2)*(4.0+sin(Pi/2))',
                                              FLEUR_DEFINED_CONSTANTS)) == -37.94528049
    assert pytest.approx(calculate_expression('(2+5)**(1/2.0)', FLEUR_DEFINED_CONSTANTS)) == 2.64575
    assert pytest.approx(calculate_expression('-0.5**(4%(2.0+1.0))', FLEUR_DEFINED_CONSTANTS)) == -0.5


def test_calculate_expression_errors():

    with pytest.raises(ValueError, match=r'Invalid expression: Expected Bracket after function name'):
        calculate_expression('sin1.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid expression: log\(x\), x\<\=0'):
        calculate_expression('log(0.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid expression: sqrt\(x\), x\<0'):
        calculate_expression('sqrt(-1.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid expression: asin\(x\), \|x\|\>1'):
        calculate_expression('asin(2.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid expression: acos\(x\), \|x\|\>1'):
        calculate_expression('acos(2.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Unknown string expression: A'):
        calculate_expression('(3.0 + A)*5', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid Expression: Found operator / in the beginning of expression'):
        calculate_expression('/1.0', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid Expression: Operator following operator'):
        calculate_expression('(3.0 + /1.0)*5', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Undefined Expression: Division by zero'):
        calculate_expression('1.0/(3.0-3.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Undefined Expression: 0\^0'):
        calculate_expression('0.0^(3.0-3.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Undefined Expression: x\^y, x\<0 and y not integer'):
        calculate_expression('-1.0^(1.0-1.0/2.0)', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid expression: Found unexpected character &'):
        calculate_expression('1.0&4', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Cannot parse number: Found two decimal points'):
        calculate_expression('1..09', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError, match=r'Invalid Expression: Unbalanced parentheses'):
        calculate_expression('(1.04+7.89*(6.3/Pi)', FLEUR_DEFINED_CONSTANTS)



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Here basic functionality is provided for setting default parameters for plotting
and ensuring consitent values for these
"""

import copy
from functools import wraps
from contextlib import contextmanager
from collections import ChainMap
import warnings


@contextmanager
def NestedPlotParameters(plotter_object):
    """
    Contextmanager for nested plot function calls
    Will reset function defaults and parameters to previous
    values after exiting

    :param plotter_object: Plotter instance
    """
    assert isinstance(plotter_object, Plotter), \
           'The NestedPlotParameters contextmanager should only be used for Plotter objects'

    function_defaults_before = copy.deepcopy(plotter_object._params.maps[2])
    parameters_before = copy.deepcopy(plotter_object._params.maps[0])

    try:
        yield
    finally:  #Also performed if exception is thrown??
        plotter_object._params.maps[2] = function_defaults_before
        plotter_object._params.maps[0] = parameters_before


def ensure_plotter_consistency(plotter_object):
    """
    Decorator for plot functions to ensure that the
    Parameters are reset even if an error occurs in the function
    Additionally checks are performed that the parameters are reset after execution
    and the defaults are never changed in a plot function

    :param plotter_object: Plotter instance to be checked for consistency
    """

    assert isinstance(plotter_object, Plotter), \
           'The ensure_plotter_consistency decorator should only be used for Plotter objects'

    def ensure_plotter_consistency_decorator(func):
        """
        Decorator that adds checks on the Plotter object
        """

        @wraps(func)
        def ensure_consistency(*args, **kwargs):
            """
            If an error is encountered in the decorated function the parameters
            of the plotter object are reset to avoid unintended sideeffects

            Also after execution the defaults and parameters are checked to make sure
            they are consistent
            """

            global_defaults_before = copy.deepcopy(plotter_object._params.maps[1])

            try:
                res = func(*args, **kwargs)
            except Exception:
                plotter_object.remove_added_parameters()
                plotter_object.reset_parameters()
                plotter_object._params.maps[2] = {}
                raise  #We do not want to erase the exception only wedge in the call to reset_parameters
            else:
                plotter_object.remove_added_parameters()
                plotter_object.reset_parameters()
                plotter_object._params.maps[2] = {}

            if plotter_object._params.maps[1] != global_defaults_before:
                #Reset the changes
                plotter_object._params.maps[1] = global_defaults_before
                plotter_object.remove_added_parameters()
                plotter_object.reset_parameters()
                plotter_object._params.maps[2] = {}
                raise ValueError(f"Defaults have changed inside the plotting function '{func.__name__}'")

            return res

        return ensure_consistency

    return ensure_plotter_consistency_decorator


def _generate_plot_parameters_table(defaults, descriptions):
    """
    Generate a table for the plotting parameters for the docstrings

    :param defaults: dict/chainmap with the defined defaults
    :param descriptions: dict with the description of the keys in defaults
    """

    #yapf: disable
    table = [
        '.. list-table:: Plot Parameters',
        '       :widths: 15 60 25',
        '       :header-rows: 1',
        '       :class: tight-table',
        '',
        '       * - Name',
        '         - Description',
        '         - Default value'
    ]

    for key, value in defaults.items():
        if value is None:
            value = 'No Default'
        elif not isinstance(value, dict):
            value = f'``{value}``'

        descr = descriptions.get(key, 'No Description available')

        table.extend([f'       * - ``{key}``',
                      f'         - {descr}',
                      f'         - {value}'])

    table.append('')
    #yapf: enable

    return '\n'.join(table)


class Plotter(object):
    """
    Base class for handling parameters for plotting methods. For different plotting backends
    a subclass can be created to represent the specific parameters of the backend.

    Args:
        :param default_parameters: dict with hardcoded default parameters
        :param general_keys: set of str optional, defines parameters which are
                             not allowed to change for each entry in the plot data

    Kwargs in the __init__ method are forwarded to :py:func:`Plotter.set_defaults()`
    to change the current defaults away from the hardcoded parameters.

    The Plotter class creates a hierachy of dictionaries for lookups on this object
    utilizing the `ChainMap` from the `collections` module.

    The hierachy is as follows (First entries take precedence over later entries):
        - `parameters`: set by :py:func:`~Plotter.set_parameters()` (usually arguments passed into function)
        - `user defaults`: set by :py:func:`~Plotter.set_defaults()`
        - `function defaults`: set by :py:func:`~Plotter.set_defaults()` with `default_type='function'`
        - `global defaults`: Hardcoded as fallback

    Only the `parameters` can represent parameters for multiple sets of plot calls.
    All others are used as fallback for specifying non-specified values for single plots

    The current parameters can be accessed by bracket indexing the class. A example of this is shown below.

    .. code-block:: python

        parameter_dict = {'fontsize': 16, 'linestyle': '-'}

        params = Plotter(parameter_dict)

        #Accessing a parameter
        print(params['fontsize']) # 16

        #Modifying a parameter
        params['fontsize'] = 20
        print(params['fontsize']) # 20

        #Creating a parameter set for multiple plots

        #1. Set the properties to the correct values
        params.single_plot = False
        params.num_plots = 3

        #2. Now we can set a property either by providing a list or a integer indexed dict
        #   Both of the following examples set the linestyle of the second and third plot to '--'
        params['linestyle'] = [None, '--', '--']
        params['linestyle'] = {1: '--', 2: '--'}

        # Not specified values are replaced with the default value for a single plot
        print(params['linestyle']) # ['-', '--', '--']

        #In lists properties can also be indexed via tuples
        print(params[('linestyle', 0)]) # '-'
        print(params[('linestyle', 1)]) # '--'

        #Changes to the parameters and properties are reset
        params.reset_parameters()

        print(params['linestyle']) # '-'

    """

    def __init__(self, default_parameters, general_keys=None, key_descriptions=None, **kwargs):

        self._PLOT_DEFAULTS = copy.deepcopy(default_parameters)

        #ChainMap with three dictionaries on top
        # 1. function parameters
        # 2. global defaults
        # 3. function defaults
        # 4. Hardcoded defaults
        self._params = ChainMap({}, {}, {}, self._PLOT_DEFAULTS)

        self._single_plot = True
        self._num_plots = 1
        self._added_parameters = set()

        self._GENERAL_KEYS = set()
        if general_keys is not None:
            self._GENERAL_KEYS = general_keys

        self._DESCRIPTIONS = {}
        if key_descriptions is not None:
            self._DESCRIPTIONS = key_descriptions

        if kwargs:
            self.set_defaults(continue_on_error=True, **kwargs)

    def __getitem__(self, indices):
        """
        Get the current value for the key

        :param indices: either str (specifies the key) or
                        tuple of str and int (specifies the key and index to access)

        :returns: the current parameter for the given specification. If tuple is given
                  and the paramter is a list the second item is used for the list index
        """
        if isinstance(indices, tuple):
            if len(indices) != 2:
                raise ValueError('Only Key or (Key, Index) Indexing supported!')
            key, index = indices
        else:
            key = indices
            index = None

        try:
            value = self._params[key]
            if isinstance(value, list):
                if index is None:
                    return value
                if index < len(value):
                    return value[index]
                else:
                    return value[0]
            else:
                return value
        except KeyError:
            return None

    def get_multiple_kwargs(self, keys, ignore=None):
        """
        Get multiple parameters and return them in a dictionary

        :param keys: set of keys to process
        :param ignore: str or list of str (optional), defines keys to ignore in the creation of the dict
        """

        keys_used = copy.deepcopy(keys)

        if ignore is not None:
            if not isinstance(ignore, list):
                ignore = [ignore]
            for key in ignore:
                keys_used.discard(key)

        ret_dict = {}
        for key in keys_used:
            if self[key] is not None:
                ret_dict[key] = self[key]

        return ret_dict

    @staticmethod
    def dict_of_lists_to_list_of_dicts(dict_of_lists, single_plot, num_plots):
        """
        Converts dict of lists and single values to list of length num_plots
        or single dict for single_plot=True

        :param dict_of_lists: dict to be converted
        :param single_plot: boolean, if True only a single parameter set is allowed
        :param num_plots: int of the number of allowed plots

        :returns: list of dicts
        """
        any_list = any(isinstance(val, (list, tuple)) for val in dict_of_lists.values())

        #Make sure that every entry is actually a list
        if any_list:
            for key, val in dict_of_lists.items():
                if not isinstance(val, (list, tuple)):
                    dict_of_lists[key] = [val] * num_plots
        elif not single_plot:
            dict_of_lists = {key: [value] for key, value in dict_of_lists.items()}

        list_of_dicts = dict_of_lists  #For single plot these are equivalent
        if not single_plot:
            list_of_dicts = [{key: value[index]
                              for key, value in dict_of_lists.items()}
                             for index in range(max(map(len, dict_of_lists.values())))]
            if len(list_of_dicts) != num_plots:
                if len(list_of_dicts) == 1:
                    list_of_dicts = [copy.deepcopy(list_of_dicts[0]) for i in range(num_plots)]
                else:
                    raise ValueError('Length does not match number of plots')

        return list_of_dicts

    @staticmethod
    def convert_to_complete_list(given_value, single_plot, num_plots, default=None, key=''):
        """
        Converts given value to list with length num_plots with None for the non-specified values

        :param given_value: value passed in, for multiple plots either list or dict with integer keys
        :param single_plot: bool, if True only a single parameter is allowed
        :param num_plots: int, if single_plot is False this defines the number of plots
        :param default: default value for unspecified entries
        :param key: str of the key to process
        """

        if not isinstance(given_value, dict) and not isinstance(given_value, list):
            return given_value

        ret_value = copy.copy(given_value)
        if isinstance(given_value, dict) and all(isinstance(key, int) for key in given_value):
            if single_plot:
                raise ValueError(f"Got dict with integer indices for '{key}' but only a single plot is allowed")
            #Convert to list with defaults for not specified keys
            ret_value = [ret_value[indx] if indx in ret_value else None for indx in range(num_plots)]

        if isinstance(ret_value, list):
            if single_plot:
                raise ValueError(f"Got list for key '{key}' but only a single plot is allowed")

            if len(ret_value) != num_plots:
                ret_value = ret_value.copy() + [None] * (num_plots - len(ret_value))
            ret_value = [val if val is not None else default for val in ret_value]

        return ret_value

    def set_single_default(self, key, value, default_type='global'):
        """
        Set default value for a single key/value pair

        :param key: str of the key to set
        :param value: value to set the key to
        :default_type: either 'global' or 'function'. Specifies, whether to
                       set the global defaults (not reset after function)
                       or the function defaults
        """
        if key not in self._params:
            raise KeyError(f'Unknown parameter: {key}')

        if default_type == 'global':
            self.__update_map(self._params.parents, key, value)
        elif default_type == 'function':
            self.__update_map(self._params.parents.parents, key, value)

    def __setitem__(self, key, value):
        """
        Set the given key value pair on the `Plotter._params` ChainMap
        (Always to the top layer).

        Unknown keys are forbidden.
        Keys allowed for multiple plot sets are converted to complete
        lists

        :param key: key to update
        :param value: value to use for updating
        """

        if key not in self._params:
            raise KeyError(f'Unknown parameter: {key}')

        if key not in self._GENERAL_KEYS:
            value = self.convert_to_complete_list(value,
                                                  self.single_plot,
                                                  self.num_plots,
                                                  default=self._params.parents[key],
                                                  key=key)

        self.__update_map(self._params, key, value)

    @staticmethod
    def __update_map(map_to_change, key, value):
        """
        Updates the given map with the given key value pair
        If the value is a dict it will be merged

        :param map_to_change: Mapping to change
        :param key: key to change
        :param value: value for updating the key

        """
        if isinstance(map_to_change[key], dict):
            dict_before = copy.deepcopy(map_to_change[key])
            if not isinstance(value, dict):
                if isinstance(value, list):
                    map_to_change[key] = dict_before
                else:
                    raise ValueError(f"Expected a dict for key {key} got '{value}'")
            else:
                dict_before.update(value)
                map_to_change[key] = dict_before
        else:
            map_to_change[key] = value

    def set_defaults(self, continue_on_error=False, default_type='global', **kwargs):
        """
        Set the current defaults. This method will only work if the parameters
        are not changed from the defaults. Otherwise a error is raised. This is because
        after changing the defaults the changes will be propagated to the parameters to ensure
        consistency.

        :param continue_on_error: bool, if True unknown key are simply skipped
        :default_type: either 'global' or 'function'. Specifies, whether to
                       set the global defaults (not reset after function)
                       or the function defaults

        Kwargs are used to set the defaults.
        """

        kwargs_unprocessed = copy.deepcopy(kwargs)
        if default_type == 'global':
            defaults_before = copy.deepcopy(self._params.maps[1])
        elif default_type == 'function':
            defaults_before = copy.deepcopy(self._params.maps[2])

        for key, value in kwargs.items():

            try:
                self.set_single_default(key, value, default_type=default_type)
                kwargs_unprocessed.pop(key)
            except KeyError as err:
                if not continue_on_error:
                    if default_type == 'global':
                        self._params.maps[1] = defaults_before
                    elif default_type == 'function':
                        self._params.maps[2] = defaults_before
                    raise KeyError(f'Unknown parameter: {key}') from err

        if 'extra_kwargs' in kwargs_unprocessed:
            extra_kwargs = kwargs_unprocessed.pop('extra_kwargs')
            kwargs_unprocessed.update(extra_kwargs)

        return kwargs_unprocessed

    def set_parameters(self, continue_on_error=False, **kwargs):
        """
        Set the current parameters.

        :param continue_on_error: bool, if True unknown key are simply skipped and returned

        Kwargs are used to set the defaults.
        """
        params_before = copy.deepcopy(self._params.maps[0])
        kwargs_unprocessed = copy.deepcopy(kwargs)

        for key, value in kwargs.items():
            try:
                self[key] = value
                kwargs_unprocessed.pop(key)
            except KeyError:
                if not continue_on_error:
                    self._params.maps[0] = params_before
                    raise

        if 'extra_kwargs' in kwargs_unprocessed:
            extra_kwargs = kwargs_unprocessed.pop('extra_kwargs')
            kwargs_unprocessed.update(extra_kwargs)

        return kwargs_unprocessed

    def add_parameter(self, name, default_from=None, default_val=None):
        """
        Add a new parameter to the parameters dictionary.

        :param name: str name of the parameter
        :param default_from: str (optional), if given a entry is created in the curent defaults
                             with the name and the default value of the key `default_from`

        """

        if default_val is not None:
            if default_from is not None:
                raise ValueError('Default value speciefied via default_val and default_from. Please choose one option')
        elif default_from is not None:
            default_val = self._params.parents[default_from]
            if isinstance(default_val, (dict, list)):
                default_val = copy.deepcopy(default_val)

        self._added_parameters.add(name)
        self._params.maps[2][name] = default_val

    def remove_added_parameters(self):
        """
        Remove the parameters added via :py:func:`Plotter.add_parameter()`
        """

        for key in copy.deepcopy(self._added_parameters):
            self._params.maps[2].pop(key, None)
            self._params.maps[0].pop(key, None)

    def reset_defaults(self):
        """
        Resets the defaults to the hardcoded defaults in _PLOT_DEFAULTS. Will check beforehand
        if the parameters or properties differ from the defaults and will raise an error if this
        is the case
        """
        self._params = ChainMap({}, {}, {}, self._PLOT_DEFAULTS)

    def reset_parameters(self):
        """
        Reset the parameters to the current defaults. The properties single_plot and num_plots
        are also set to default values
        """

        self._params.maps[0] = {}
        #Reset number of plots properties
        self.single_plot = True
        self.num_plots = 1

    def get_dict(self):
        """
        Return the dictionary of the current defaults. For use of printing
        """
        return dict(self._params)

    def get_description(self, key):
        """
        Get the description of the given key

        :param key: str of the key, for which the description should be printed
        """

        if key in self._DESCRIPTIONS:
            print(f'{key}:\n\n{self._DESCRIPTIONS[key]}')
        elif key in self._params:
            print(f'{key}:\n\nNo Description available')
        else:
            warnings.warn(f'{key} is not a known parameter')

    @property
    def single_plot(self):
        """
        Boolean property if True only a single Plot parameter set is allowed
        """
        return self._single_plot

    @single_plot.setter
    def single_plot(self, boolean_value):
        """
        Setter for single_plot property
        """
        self._single_plot = boolean_value

    @property
    def num_plots(self):
        """
        Integer property for number of plots produced
        """
        return self._num_plots

    @num_plots.setter
    def num_plots(self, int_value):
        """
        Setter for num_plots property
        """
        self._num_plots = int_value



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), 2018 Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.    #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
#                                                                             #
###############################################################################
"""
:author: Philipp Ruessmann
:date: 2018-11-26

Plotting utility to visualize the output of the voronoi code.

Reads files 'vertices.dat' to extract the vertices of the shapefunctions and 'positions.dat' (done in 'read_shapefun' function)
Then creates a simple matplotlib image to show the shapefunctions using the 'plot_shapefun' function.
"""
from __future__ import print_function

from __future__ import absolute_import
from numpy import array, shape
import sys
from sys import argv
from masci_tools.io.kkr_read_shapefun_info import read_shapefun
from masci_tools.vis.kkr_plot_shapefun import plot_shapefun  #, change_zoom, zoom_in
from matplotlib.pyplot import show

mode = 'all'  # 'single' or 'all'

if len(argv) > 1:
    mode = argv[1]

if mode not in ['all', 'single']:
    print('ERROR: illegal mode chosen. Only all or single can be used')
    sys.exit()

pos, vertices = read_shapefun('.')
ax = plot_shapefun(pos, vertices, mode)

# this can be used to
#if mode=='all':
#change_zoom(ax, 1.5, center=[0,0,0])
#zoom_in(ax, 3, pos, zoom_range=1)

show()



# -*- coding: utf-8 -*-
"""
Tests for the case insensitive containers implemented in util/case_insensitive_dict.py
"""
import pytest

TEST_DICT = {'ABC': 123, 'SUBDICT': {'TEST': 'test_string'}, 'miXedCase4': 5}
EXPECTED_REPR = "CaseInsensitiveDict({'abc': 123, 'subdict': {'TEST': 'test_string'}, 'mixedcase4': 5})"
EXPECTED_REPR_UPPER = "CaseInsensitiveDict({'ABC': 123, 'SUBDICT': {'TEST': 'test_string'}, 'MIXEDCASE4': 5})"

TEST_LOOKUPS = [('ABC', 123), ('abc', 123), ('aBc', 123), ('sUbDiCT', {'TEST': 'test_string'}), ('MIXEDCASE4', 5)]
TEST_LOOKUPS_NOT_EXISTENT = ['not_existent', 'MIXEDCASE$', 'AB_C']


@pytest.mark.parametrize('key, result', TEST_LOOKUPS)
def test_case_insensitive_dict_get(key, result):
    """
   Test case insensitive lookup in CaseInsensitiveDict
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    d = CaseInsensitiveDict(TEST_DICT)

    assert key in d
    assert d[key] == result
    assert repr(d) == EXPECTED_REPR


@pytest.mark.parametrize('key, result', TEST_LOOKUPS)
def test_case_insensitive_dict_upper_get(key, result):
    """
   Test case insensitive lookup in CaseInsensitiveDict with upper() as normalizing function
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    d = CaseInsensitiveDict(TEST_DICT, upper=True)

    assert key in d
    assert d[key] == result
    assert repr(d) == EXPECTED_REPR_UPPER


@pytest.mark.parametrize('key', TEST_LOOKUPS_NOT_EXISTENT)
def test_case_insensitive_dict_get_not_existent(key):
    """
   Test case insensitive lookup (non existent keys) in CaseInsensitiveDict
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    d = CaseInsensitiveDict(TEST_DICT)

    assert key not in d
    with pytest.raises(KeyError):
        val = d[key]


@pytest.mark.parametrize('key', TEST_LOOKUPS_NOT_EXISTENT)
def test_case_insensitive_dict_upper_get_not_existent(key):
    """
   Test case insensitive lookup (non existent keys) in CaseInsensitiveDict with upper() as normalizing function
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    d = CaseInsensitiveDict(TEST_DICT, upper=True)

    assert key not in d
    with pytest.raises(KeyError):
        val = d[key]


@pytest.mark.parametrize('key, result', TEST_LOOKUPS)
def test_case_insensitive_dict_pop(key, result):
    """
   Test case insensitive pop in CaseInsensitiveDict
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    d = CaseInsensitiveDict(TEST_DICT)

    assert key in d
    assert d.pop(key) == result
    assert key not in d


@pytest.mark.parametrize('key, result', TEST_LOOKUPS)
def test_case_insensitive_dict_pop_upper(key, result):
    """
   Test case insensitive pop in CaseInsensitiveDict with upper() as normalizing function
   """
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict

    d = CaseInsensitiveDict(TEST_DICT, upper=True)

    assert key in d
    assert d.pop(key) == result
    assert key not in d



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module defines decorators for the ParseTasks class to make extending/modifying the parser
more convenient

Up till now 3 decorators are defined:
    - ```register_migration``` marks a function of making backwards incompatible changes
      to the parsing tasks
    - ```register_parsing_function``` gives a mappimg between available parsing functions
      and the keywords in the parsing tasks
    - ```conversion_function``` makes the decorated function available to be called easily
      after a certain parsing task has occured
"""
from masci_tools.util.parse_tasks import ParseTasks
from functools import wraps


def register_migration(base_version, target_version):
    """
    Decorator to add migration for task definition dictionary to the ParseTasks class
    The function should only take the dict of task definitions as an argument

    :param base_version: str of the version, from which the migration starts
    :param target_version: str or list of str with the versions that work after
                           the migration has been performed

    """

    def migration_decorator(func):
        """
        Return decorated ParseTasks object with _migrations dict attribute
        Here all registered migrations are inserted
        """

        @wraps(func)
        def migration(*args):
            """Decorator for migration function"""
            return func(*args)

        if not hasattr(ParseTasks, '_migrations'):
            ParseTasks._migrations = {}  # pylint: disable=protected-access
        if not base_version in ParseTasks._migrations:
            ParseTasks._migrations[base_version] = {}

        target_version_list = target_version
        if not isinstance(target_version_list, list):
            target_version_list = [target_version_list]
        for valid_version in target_version_list:
            ParseTasks._migrations[base_version][valid_version] = migration  # pylint: disable=protected-access

            for valid_version_2 in target_version_list:
                if valid_version == valid_version_2:
                    continue
                if int(valid_version.split('.')[1]) > int(valid_version_2.split('.')[1]):
                    if valid_version not in ParseTasks._migrations:
                        ParseTasks._migrations[valid_version] = {}
                    ParseTasks._migrations[valid_version][valid_version_2] = 'compatible'
                else:
                    if valid_version_2 not in ParseTasks._migrations:
                        ParseTasks._migrations[valid_version_2] = {}
                    ParseTasks._migrations[valid_version_2][valid_version] = 'compatible'

        return migration

    return migration_decorator


def register_parsing_function(parse_type_name, all_attribs_keys=False):
    """
    Decorator to add parse type for task definition dictionary.

    :param parse_type_name: str, the function can be selected in task defintions
                            via this string
    :param all_attribs_keys: bool, if True the arguments for parsing multiple attributes
                             are valid

    The decorated function has to have the following arguments:
        :param node: etree Element, on which to execute the xpath evaluations
        :param schema_dict: dict, containing all the path information and more
        :param name: str, name of the tag/attribute
        :param parser_info_out: dict, with warnings, info, errors, ...
        :param kwargs: here all other keyword arguments are collected

    """

    def parse_type_decorator(func):
        """
        Return decorated ParseTasks object with _parse_functions dict attribute
        Here all registered migrations are inserted
        """

        @wraps(func)
        def parse_type(*args, **kwargs):
            """Decorator for parse_type function"""
            return func(*args, **kwargs)

        if not hasattr(ParseTasks, '_parse_functions'):
            ParseTasks._parse_functions = {}  # pylint: disable=protected-access
            ParseTasks._all_attribs_function = set()

        ParseTasks._parse_functions[parse_type_name] = parse_type  # pylint: disable=protected-access
        if all_attribs_keys:
            ParseTasks._all_attribs_function.add(parse_type_name)

        return parse_type

    return parse_type_decorator


def conversion_function(func):
    """
    Marks a function as a conversion function, which can be called after
    performing a parsing task. The function can be specified via the _conversions
    control key in the task definitions.

    A conversion function has to have the following arguments:
        :param out_dict: dict with the previously parsed information
        :param parser_info_out: dict, with warnings, info, errors, ...

    and return only the modified output dict
    """

    @wraps(func)
    def convert_func(*args, **kwargs):
        """Decorator for parse_type function"""
        return func(*args, **kwargs)

    if not hasattr(ParseTasks, '_conversion_functions'):
        ParseTasks._conversion_functions = {}  # pylint: disable=protected-access

    ParseTasks._conversion_functions[func.__name__] = convert_func  # pylint: disable=protected-access

    return convert_func



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Load all fleur schema related functions
"""

from .inpschema_todict import *
from .outschema_todict import *
from .add_fleur_schema import *
from .schema_dict import *

__all__ = [
    'create_inpschema_dict', 'create_outschema_dict', 'add_fleur_schema', 'InputSchemaDict', 'OutputSchemaDict',
    'schema_dict_version_dispatch'
]



# -*- coding: utf-8 -*-
"""
Tests of the Lockable containers
"""
import pytest

TEST_DICT = {'A': 3, 'B': {'Test': [], 'Test2': 'A'}}
TEST_LIST = [3, {'Test': [], 'Test2': 'A'}]


def test_lockable_dict():
    """
   test the LockableDict behaviour
   """
    from masci_tools.util.lockable_containers import LockableDict, LockableList

    l = LockableDict(TEST_DICT)

    l['test_set'] = 'iamallowedtochange'

    assert l['test_set'] == 'iamallowedtochange'

    l.freeze()

    assert l._locked
    assert isinstance(l['B'], LockableDict)
    assert l['B']._locked
    assert isinstance(l['B']['Test'], LockableList)
    assert l['B']['Test']._locked

    with pytest.raises(RuntimeError, match='Modification not allowed'):
        l['test_set'] = 'notanymore'

    with pytest.raises(RuntimeError, match='Modification not allowed'):
        l.pop('B')


def test_lockable_list():
    """
   test the LockableList behaviour
   """
    from masci_tools.util.lockable_containers import LockableDict, LockableList

    l = LockableList(TEST_LIST)
    print([type(val) for val in l])

    l.append('iamallowedtochange')

    assert l[-1] == 'iamallowedtochange'

    l.freeze()

    assert l._locked
    assert isinstance(l[1], LockableDict)
    assert l[1]._locked
    assert isinstance(l[1]['Test'], LockableList)
    assert l[1]['Test']._locked

    with pytest.raises(RuntimeError, match='Modification not allowed'):
        l[2] = 'notanymore'

    with pytest.raises(RuntimeError, match='Modification not allowed'):
        l.append('B')

    with pytest.raises(RuntimeError, match='Modification not allowed'):
        l.pop()


def test_lock_container_contextmanager():
    """
   Test the LockContainer contextmanager
   """
    from masci_tools.util.lockable_containers import LockableDict, LockableList, LockContainer

    l_dict = LockableDict(TEST_DICT)
    l_list = LockableList(TEST_LIST)

    with LockContainer(l_dict):
        l_list.append('test')
        assert l_list[-1] == 'test'
        with pytest.raises(RuntimeError, match='Modification not allowed'):
            l_dict['Test2'] = {'subdict': 4}

    assert not l_dict._locked
    assert l_dict['B'].pop('Test2') == 'A'

    with LockContainer(l_list):
        with pytest.raises(RuntimeError, match='Modification not allowed'):
            l_list.append('test')
        l_dict['Test2'] = {'subdict': 4}

    assert not l_list._locked
    l_list[1]['A2'] = 'A3'

    l_dict.freeze()

    with pytest.raises(AssertionError, match='LockableDict was already locked before entering the contextmanager'):
        with LockContainer(l_dict):
            pass

    with pytest.raises(AssertionError, match='Wrong type Got:'):
        with LockContainer([]):
            pass



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
from matplotlib import cm


def FSqdos2D(p0='./',
             totonly=True,
             s=20,
             ls_ef=':',
             lw_ef=1,
             color='',
             reload_data=False,
             clrbar=True,
             atoms=[],
             ax=None,
             nosave=False,
             noalat=False,
             cmap=cm.jet,
             noplot=False,
             return_data=False,
             pclrmesh=False,
             logscale=True,
             ef=None):
    """ plotting routine for dos files """
    # import dependencies
    import numpy as np
    import matplotlib.pyplot as plt
    from os import listdir
    from os.path import isdir
    from subprocess import check_output

    # deal with input of file handle instead of path (see plot_kkr of aiida_kkr)
    if not isinstance(p0, str):
        pathname_with_file = p0.name
        p0 = pathname_with_file.replace('/qdos.01.1.dat', '')  #dos.atom1

    # read in data
    if p0[-1] != '/':
        p0 += '/'

    # read EF if not given as input
    if ef is None:
        if 'potential' in listdir(p0):
            ef = float(open(p0 + 'potential').readlines()[3].split()[1])
        else:
            ef = 0

    if noalat:
        a0 = 1.
        alat = 1.
    else:
        alat = float(
            check_output('grep ALATBASIS ' + p0 + 'inputcard', shell=True).decode('utf-8').split('=')[1].split()[0])
        a0 = 2 * np.pi / alat / 0.52918

    if reload_data or 'saved_data_qdos.npy' not in np.sort(listdir(p0)):
        first = True
        print('reading qdos')
        j = 0
        for i in np.sort(listdir(p0)):
            if 'qdos.' in i[:6] and not isdir(p0 + i):
                j += 1
                iatom = int(i.replace('qdos.', '').split('.')[0])
                if atoms == [] or iatom in atoms:
                    tmp = np.loadtxt(p0 + i)
                    tmp[:, 2:5] = tmp[:, 2:5] * a0
                    print(i, iatom)
                    if first:
                        d = tmp
                        first = False
                    else:
                        d[:, 5:] += tmp[:, 5:]
        if not nosave:
            np.save(p0 + 'saved_data_qdos', d)
    else:
        d = np.load(p0 + 'saved_data_qdos.npy')

    xlab = r'kx'
    ylab = r'ky'
    if a0 != 1.:
        xlab = r'$k_x (\AA^{-1})$'
        ylab = r'$k_y (\AA^{-1})$'

    # plot dos
    data = np.abs(np.sum(d[:, 5:], axis=1))
    if logscale:
        data = np.log(data)
    x, y = d[:, 2], d[:, 3]

    if not noplot:
        if ax is None:
            if pclrmesh:
                lx = len(set(x))
                ly = len(set(y))
                x = np.linspace(x.min(), x.max(), lx)
                y = np.linspace(y.min(), y.max(), ly)
                plt.pcolormesh(x, y, data.reshape(lx, ly).T, cmap=cmap)
            else:
                plt.scatter(x, y, c=data, s=s, lw=0, cmap=cmap)
            if clrbar:
                plt.colorbar()
        else:
            ax.scatter(x, y, c=data, lw=0, cmap=cmap)

        # set axis labels
        if ax is None:
            plt.xlabel(xlab)
            plt.ylabel(ylab)
            plt.axis('equal')

    if return_data:
        return x, y, data



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Small utility functions for inspecting hdf files and converting the
complete file structure into a python dictionary
"""

import h5py
import numpy as np


def hdfList(name, obj):
    """
    Print the name of the current object (indented to create a nice tree structure)

    Also prints attribute values and dataset shapes and datatypes
    """
    print(f"{name.split('/')[-1]:>{len(name)-1}}: {type(obj)}")

    ref_length = len(name) - len(name.split('/')[-1]) + 4

    if isinstance(obj, h5py.Dataset):
        print(f"{'Datatype:':>{ref_length+9}} {obj.dtype}")
        print(f"{'Shape:':>{ref_length+6}} {obj.shape}\n")

    if obj.attrs:
        print(f"{'Attributes:':>{ref_length+9}}")
        for attr_name, attr_val in obj.attrs.items():
            if len(attr_val) == 1:
                attr_val = attr_val[0]
            print(f'{attr_name:>{ref_length+len(attr_name)}}: {attr_val}')
        print('')


def h5dump(file, group='/'):
    """
    Shows the overall filestructure of an hdf file
    Goes through all groups and subgroups and prints the attributes
    or the shape and datatype of the datasets

    :param filepath: path to the hdf file
    """
    with h5py.File(file, 'r') as file_hdf:
        if group != '/':
            print(f'Starting from path {group}')
            hdfList(group, file_hdf[group])
            print('This path contains: \n')
        file_hdf[group].visititems(hdfList)


def read_hdf_simple(file, flatten=False):
    """
    Reads in an hdf file and returns its context in a nested dictionary

    :param filepath: path or filehandle to the hdf file
    :param flatten: bool, if True the dictionary will be flattened (does not check for lost information)

    :returns: two dictionaries, one with the datasets the other
              with the attributes in the file

    **Non unique group attribute or dataset names will be overwritten in the return dict**
    """

    datasets = {}
    group_attrs = {}

    with h5py.File(file, 'r') as file_hdf:
        datasets, group_attrs = read_groups(file_hdf, flatten=flatten)

    return datasets, group_attrs


def read_groups(hdfdata, flatten=False):
    """
    Recursive function to read a hdf datastructure and extract the datasets
    and attributes

    :param hdfdata: current hdf group to process
    :param flatten: bool, if True the dictionary will be flattened (does not check for lost information)

    :returns: two dictionaries, one with the datasets the other
              with the attributes in the file
    """
    datasets = {}
    attrs = {}

    for name, attr_val in hdfdata.attrs.items():
        if len(attr_val) == 1:
            attr_val = attr_val[0]
        attrs[name] = attr_val

    for key, val in hdfdata.items():

        if isinstance(val, h5py.Dataset):
            datasets[key] = np.array(val)
        else:
            new_datasets, new_attrs = read_groups(val, flatten=flatten)
            if not flatten:
                if new_datasets:
                    datasets[key] = new_datasets
                if new_attrs:
                    attrs[key] = new_attrs
            else:
                datasets.update(new_datasets)
                attrs.update(new_attrs)

    return datasets, attrs



# -*- coding: utf-8 -*-
"""
Test of the utility functions for the schema dictionaries
both path finding and easy information extraction
"""
import pytest
import os
import numpy as np
from masci_tools.io.parsers.fleur.fleur_schema import InputSchemaDict, OutputSchemaDict
from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
from pprint import pprint
import logging

LOGGER = logging.getLogger(__name__)

#Load different schema versions (for now only input schemas)
schema_dict_34 = InputSchemaDict.fromVersion('0.34')
schema_dict_27 = InputSchemaDict.fromVersion('0.27')
schema_dict_31 = InputSchemaDict.fromVersion('0.31')
outschema_dict_34 = OutputSchemaDict.fromVersion('0.34')
outschema_dict_31 = OutputSchemaDict.fromVersion('0.31')

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
TEST_INPXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/inp2.xml')
TEST_OUTXML_PATH = os.path.join(FILE_PATH, 'files/fleur/Max-R5/GaAsMultiUForceXML/files/out.xml')
TEST_OUTXML_PATH2 = os.path.join(FILE_PATH, 'files/fleur/Max-R5/FePt_film_SSFT_LO/files/out.xml')


def test_get_tag_xpath_input():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_tag_xpath

    #First example easy (magnetism tag is unique and should not differ between the versions)
    assert get_tag_xpath(schema_dict_27, 'magnetism') == '/fleurInput/calculationSetup/magnetism'
    assert get_tag_xpath(schema_dict_34, 'magnetism') == '/fleurInput/calculationSetup/magnetism'

    #Differing paths between the version
    assert get_tag_xpath(schema_dict_27, 'bzIntegration') == '/fleurInput/calculationSetup/bzIntegration'
    assert get_tag_xpath(schema_dict_34, 'bzIntegration') == '/fleurInput/cell/bzIntegration'

    #Non existent tag in old version
    assert get_tag_xpath(schema_dict_34, 'DMI') == '/fleurInput/forceTheorem/DMI'
    with pytest.raises(ValueError, match='The tag DMI has no possible paths with the current specification.'):
        get_tag_xpath(schema_dict_27, 'DMI')

    #Multiple possible paths
    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_tag_xpath(schema_dict_27, 'ldaU')
    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_tag_xpath(schema_dict_34, 'ldaU')


def test_get_relative_tag_xpath_input():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_relative_tag_xpath

    assert get_relative_tag_xpath(schema_dict_34, 'magnetism', 'calculationSetup') == './magnetism'
    assert get_relative_tag_xpath(schema_dict_34, 'magnetism', 'magnetism') == './'

    assert get_relative_tag_xpath(schema_dict_34, 'DMI', 'forceTheorem') == './DMI'
    with pytest.raises(ValueError, match='The tag DMI has no possible relative paths with the current specification.'):
        get_relative_tag_xpath(schema_dict_27, 'DMI', 'forceTheorem')

    with pytest.raises(ValueError,
                       match='The tag ldaU has multiple possible relative paths with the current specification.'):
        get_relative_tag_xpath(schema_dict_34, 'ldaU', 'fleurInput')

    assert get_relative_tag_xpath(schema_dict_34, 'ldaU', 'fleurInput',
                                  contains='species') == './atomSpecies/species/ldaU'
    assert get_relative_tag_xpath(schema_dict_34, 'ldaU', 'species') == './ldaU'


def test_get_relative_tag_xpath_output():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_relative_tag_xpath

    assert get_relative_tag_xpath(outschema_dict_34, 'iteration', 'scfLoop') == './iteration'
    assert get_relative_tag_xpath(outschema_dict_34, 'iteration', 'iteration') == './'

    assert get_relative_tag_xpath(outschema_dict_34, 'densityMatrixFor', 'ldaUDensityMatrix') == './densityMatrixFor'


def test_get_tag_xpath_output():
    """
    Test the path finding for tags for the output schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_tag_xpath

    #absolute
    assert get_tag_xpath(outschema_dict_34, 'iteration') == '/fleurOutput/scfLoop/iteration'
    assert get_tag_xpath(outschema_dict_31, 'iteration') == '/fleurOutput/scfLoop/iteration'

    #relative paths
    assert get_tag_xpath(outschema_dict_34, 'totalEnergy') == './totalEnergy'
    assert get_tag_xpath(outschema_dict_31, 'totalEnergy') == './totalEnergy'

    #Non existent tag
    assert get_tag_xpath(outschema_dict_34, 'fleurInput') == '/fleurOutput/fleurInput'
    with pytest.raises(ValueError, match='The tag fleurInput has no possible paths with the current specification.'):
        get_tag_xpath(outschema_dict_31, 'fleurInput')

    assert get_tag_xpath(outschema_dict_31, 'inputData') == '/fleurOutput/inputData'
    with pytest.raises(ValueError, match='The tag inputData has no possible paths with the current specification.'):
        get_tag_xpath(outschema_dict_34, 'inputData')

    #Multiple possible paths
    with pytest.raises(ValueError,
                       match='The tag spinDependentCharge has multiple possible paths with the current specification.'):
        get_tag_xpath(outschema_dict_31, 'spinDependentCharge')
    with pytest.raises(ValueError,
                       match='The tag spinDependentCharge has multiple possible paths with the current specification.'):
        get_tag_xpath(outschema_dict_34, 'spinDependentCharge')


def test_get_tag_xpath_contains():
    """
    Test the selection of paths based on a contained keyword
    """
    from masci_tools.util.schema_dict_util import get_tag_xpath

    schema_dict = schema_dict_34

    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_tag_xpath(schema_dict, 'ldaU')
    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_tag_xpath(schema_dict, 'ldaU', contains='atom')

    assert get_tag_xpath(schema_dict, 'ldaU', contains='calculationSetup') == '/fleurInput/calculationSetup/ldaU'
    assert get_tag_xpath(schema_dict, 'ldaU', contains='species') == '/fleurInput/atomSpecies/species/ldaU'
    assert get_tag_xpath(schema_dict, 'ldaU', contains='Group') == '/fleurInput/atomGroups/atomGroup/ldaU'

    with pytest.raises(ValueError, match='The tag ldaU has no possible paths with the current specification.'):
        get_tag_xpath(schema_dict, 'ldaU', contains='group')


def test_get_tag_xpath_notcontains():
    """
    Test the selection of paths based on a not contained keyword
    """
    from masci_tools.util.schema_dict_util import get_tag_xpath

    schema_dict = schema_dict_34

    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_tag_xpath(schema_dict, 'ldaU')
    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_tag_xpath(schema_dict, 'ldaU', not_contains='calculationSetup')

    assert get_tag_xpath(schema_dict, 'ldaU', not_contains='atom') == '/fleurInput/calculationSetup/ldaU'
    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        assert get_tag_xpath(schema_dict, 'ldaU', not_contains='Group') == '/fleurInput/atomSpecies/species/ldaU'

    assert get_tag_xpath(schema_dict, 'ldaU', contains='atom',
                         not_contains='species') == '/fleurInput/atomGroups/atomGroup/ldaU'


def test_get_tagattrib_xpath_case_insensitivity():
    """
    Test that the selection works with case insensitivity
    """
    from masci_tools.util.schema_dict_util import get_tag_xpath, get_attrib_xpath

    schema_dict = schema_dict_34

    assert get_tag_xpath(schema_dict, 'bzIntegration') == '/fleurInput/cell/bzIntegration'
    assert get_tag_xpath(schema_dict, 'BZINTEGRATION') == '/fleurInput/cell/bzIntegration'
    assert get_tag_xpath(schema_dict, 'bzintegration') == '/fleurInput/cell/bzIntegration'
    assert get_tag_xpath(schema_dict, 'bZInTegrAtIon') == '/fleurInput/cell/bzIntegration'

    assert get_attrib_xpath(schema_dict, 'jspins') == '/fleurInput/calculationSetup/magnetism/@jspins'
    assert get_attrib_xpath(schema_dict, 'JSPINS') == '/fleurInput/calculationSetup/magnetism/@jspins'
    assert get_attrib_xpath(schema_dict, 'jSpInS') == '/fleurInput/calculationSetup/magnetism/@jspins'


def test_get_attrib_xpath_input():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_attrib_xpath

    #First example easy (magnetism tag is unique and should not differ between the versions)
    assert get_attrib_xpath(schema_dict_27, 'jspins') == '/fleurInput/calculationSetup/magnetism/@jspins'
    assert get_attrib_xpath(schema_dict_34, 'jspins') == '/fleurInput/calculationSetup/magnetism/@jspins'

    #Differing paths between the version
    assert get_attrib_xpath(schema_dict_27, 'mode') == '/fleurInput/calculationSetup/bzIntegration/@mode'
    assert get_attrib_xpath(schema_dict_34, 'mode') == '/fleurInput/cell/bzIntegration/@mode'

    #Non existent tag in old version
    assert get_attrib_xpath(schema_dict_34, 'l_mtNocoPot',
                            exclude=['other']) == '/fleurInput/calculationSetup/magnetism/mtNocoParams/@l_mtNocoPot'
    with pytest.raises(ValueError,
                       match='The attrib l_mtNocoPot has no possible paths with the current specification.'):
        get_attrib_xpath(schema_dict_27, 'l_mtNocoPot')

    #Multiple possible paths
    with pytest.raises(ValueError,
                       match='The attrib l_amf has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict_27, 'l_amf')
    with pytest.raises(ValueError,
                       match='The attrib l_amf has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict_34, 'l_amf')


def test_get_relative_attrib_xpath_input():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_relative_attrib_xpath

    #First example easy (magnetism tag is unique and should not differ between the versions)
    assert get_relative_attrib_xpath(schema_dict_34, 'jspins', 'calculationSetup') == './magnetism/@jspins'
    assert get_relative_attrib_xpath(schema_dict_34, 'jspins', 'magnetism') == './@jspins'

    assert get_relative_attrib_xpath(schema_dict_34, 'jspins', 'magnetism', tag_name='magnetism') == './@jspins'

    with pytest.raises(ValueError, match='No attribute jspins found at tag calculationSetup'):
        get_relative_attrib_xpath(schema_dict_34, 'jspins', 'calculationSetup', tag_name='calculationSetup')

    #Non existent tag in old version
    assert get_relative_attrib_xpath(schema_dict_34, 'l_mtNocoPot', 'magnetism') == './mtNocoParams/@l_mtNocoPot'
    with pytest.raises(
            ValueError,
            match='The attrib l_mtNocoPot has multiple possible relative paths with the current specification.'):
        get_relative_attrib_xpath(schema_dict_34, 'l_mtNocoPot', 'fleurInput')

    with pytest.raises(ValueError,
                       match='The attrib l_mtNocoPot has no possible relative paths with the current specification.'):
        get_relative_attrib_xpath(schema_dict_34, 'l_mtNocoPot', 'output')

    #Multiple possible paths
    with pytest.raises(ValueError,
                       match='The attrib l_amf has multiple possible relative paths with the current specification.'):
        get_relative_attrib_xpath(schema_dict_34, 'l_amf', 'fleurInput')

    assert get_relative_attrib_xpath(schema_dict_34, 'l_amf', 'fleurInput', contains='species',
                                     not_contains='ldaHIA') == './atomSpecies/species/ldaU/@l_amf'
    assert get_relative_attrib_xpath(schema_dict_34, 'l_amf', 'ldaU') == './@l_amf'


def test_get_attrib_xpath_output():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_attrib_xpath

    #absolute
    assert get_attrib_xpath(outschema_dict_31, 'nat') == '/fleurOutput/numericalParameters/atomsInCell/@nat'
    assert get_attrib_xpath(outschema_dict_34, 'nat') == '/fleurOutput/numericalParameters/atomsInCell/@nat'

    #relative
    assert get_attrib_xpath(outschema_dict_31, 'qvectors') == './Forcetheorem_SSDISP/@qvectors'
    assert get_attrib_xpath(outschema_dict_34, 'qvectors') == './Forcetheorem_SSDISP/@qvectors'


def test_get_relative_attrib_xpath_output():
    """
    Test the path finding for tags for the input schema without additional options
    And verify with different version of the schema
    """
    from masci_tools.util.schema_dict_util import get_relative_attrib_xpath

    assert get_relative_attrib_xpath(outschema_dict_34, 'nat', 'numericalParameters') == './atomsInCell/@nat'
    assert get_relative_attrib_xpath(outschema_dict_34, 'nat', 'atomsInCell') == './@nat'

    assert get_relative_attrib_xpath(outschema_dict_34, 'qvectors', '.') == './Forcetheorem_SSDISP/@qvectors'
    assert get_relative_attrib_xpath(outschema_dict_34, 'qvectors', 'Forcetheorem_SSDISP') == './@qvectors'


def test_get_attrib_xpath_contains():
    """
    Test the selection of paths based on a contained keyword
    """
    from masci_tools.util.schema_dict_util import get_attrib_xpath

    schema_dict = schema_dict_34

    with pytest.raises(ValueError,
                       match='The attrib l_mperp has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'l_mperp')

    assert get_attrib_xpath(schema_dict, 'l_mperp',
                            contains='magnetism') == '/fleurInput/calculationSetup/magnetism/mtNocoParams/@l_mperp'
    assert get_attrib_xpath(schema_dict, 'l_mperp',
                            contains='greensFunction') == '/fleurInput/calculationSetup/greensFunction/@l_mperp'

    with pytest.raises(ValueError, match='The attrib l_mperp has no possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'l_mperp', contains='atom')


def test_get_attrib_xpath_notcontains():
    """
    Test the selection of paths based on a contained keyword
    """
    from masci_tools.util.schema_dict_util import get_attrib_xpath

    schema_dict = schema_dict_34

    with pytest.raises(ValueError,
                       match='The attrib l_mperp has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'l_mperp')

    assert get_attrib_xpath(
        schema_dict, 'l_mperp',
        not_contains='greensFunction') == '/fleurInput/calculationSetup/magnetism/mtNocoParams/@l_mperp'
    assert get_attrib_xpath(schema_dict, 'l_mperp',
                            not_contains='magnetism') == '/fleurInput/calculationSetup/greensFunction/@l_mperp'

    assert get_attrib_xpath(schema_dict, 'l_mperp', contains='greensFunction',
                            not_contains='magnetism') == '/fleurInput/calculationSetup/greensFunction/@l_mperp'

    with pytest.raises(ValueError, match='The attrib l_mperp has no possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'l_mperp', not_contains='calculationSetup')


def test_get_attrib_xpath_exclude():
    """
    Test the selection of paths based on a contained keyword
    """
    from masci_tools.util.schema_dict_util import get_attrib_xpath

    schema_dict = schema_dict_34

    with pytest.raises(ValueError,
                       match='The attrib alpha has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'alpha')

    assert get_attrib_xpath(schema_dict, 'alpha', exclude=['unique_path',
                                                           'other']) == '/fleurInput/calculationSetup/scfLoop/@alpha'
    with pytest.raises(ValueError,
                       match='The attrib alpha has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'alpha', exclude=['unique'])

    assert get_attrib_xpath(schema_dict, 'alpha', not_contains='atom',
                            exclude=['unique'
                                     ]) == '/fleurInput/calculationSetup/greensFunction/contourSemicircle/@alpha'


def test_get_attrib_xpath_exclude_output():
    """
    Test the selection of paths based on a contained keyword
    """
    from masci_tools.util.schema_dict_util import get_attrib_xpath

    schema_dict = outschema_dict_34

    with pytest.raises(ValueError,
                       match='The attrib units has multiple possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'units')

    assert get_attrib_xpath(schema_dict, 'units', contains='DMI') == './Forcetheorem_DMI/@units'
    assert get_attrib_xpath(schema_dict, 'units', exclude=['other'], contains='DMI') == './Forcetheorem_DMI/@units'

    with pytest.raises(ValueError, match='The attrib units has no possible paths with the current specification.'):
        get_attrib_xpath(schema_dict, 'units', exclude=['unique_path'], contains='DMI')


def test_get_tag_info():
    """
    Basic test of the `get_tag_info()` function
    """
    from masci_tools.util.schema_dict_util import get_tag_info

    schema_dict = schema_dict_34

    EXPECTED_RESULT = {
        'attribs': {'name', 'element', 'atomicNumber'},
        'optional': {
            'energyParameters', 'prodBasis', 'special', 'force', 'nocoParams', 'modInitDen', 'ldaU', 'ldaHIA',
            'greensfCalculation', 'torgueCalculation', 'lo'
        },
        'optional_attribs': {
            'element': None
        },
        'order': [
            'mtSphere', 'atomicCutoffs', 'electronConfig', 'energyParameters', 'prodBasis', 'special', 'force',
            'nocoParams', 'modInitDen', 'ldaU', 'ldaHIA', 'greensfCalculation', 'torgueCalculation', 'lo'
        ],
        'several': {'ldaU', 'ldaHIA', 'greensfCalculation', 'lo'},
        'simple': {
            'mtSphere', 'atomicCutoffs', 'energyParameters', 'prodBasis', 'special', 'force', 'nocoParams',
            'modInitDen', 'ldaU', 'lo'
        },
        'text':
        set(),
        'complex': {'electronConfig', 'ldaHIA', 'greensfCalculation', 'torgueCalculation'}
    }

    res, path = get_tag_info(schema_dict, 'species')

    assert res == EXPECTED_RESULT
    assert path == '/fleurInput/atomSpecies/species'

    res = get_tag_info(schema_dict, 'species', path_return=False)

    assert res == EXPECTED_RESULT

    with pytest.raises(ValueError, match='The tag ldaHIA has multiple possible paths with the current specification.'):
        res = get_tag_info(schema_dict, 'ldaHIA')

    EXPECTED_RESULT = {
        'attribs': {'l', 'U', 'J', 'phi', 'theta', 'l_amf', 'init_occ', 'kkintgrCutoff', 'label'},
        'optional': {'exc', 'cFCoeff', 'addArg'},
        'optional_attribs': {
            'phi': '0.0',
            'theta': '0.0',
            'init_occ': 'calc',
            'kkintgrcutoff': 'calc',
            'label': 'default'
        },
        'order': ['exc', 'cFCoeff', 'addArg'],
        'several': {'exc', 'cFCoeff', 'addArg'},
        'simple': {'exc', 'cFCoeff', 'addArg'},
        'text': set(),
        'complex': set()
    }

    res, path = get_tag_info(schema_dict, 'ldaHIA', contains='species')

    assert res == EXPECTED_RESULT
    assert path == '/fleurInput/atomSpecies/species/ldaHIA'

    EXPECTED_RESULT = {
        'attribs': {
            'itmaxHubbard1', 'beta', 'minoccDistance', 'minmatDistance', 'n_occpm', 'dftspinpol', 'fullMatch',
            'l_nonsphDC', 'l_correctEtot', 'l_forceHIAiteration'
        },
        'optional': set(),
        'optional_attribs': {
            'beta': '100.0',
            'minoccdistance': '0.01',
            'minmatdistance': '0.001',
            'n_occpm': '2',
            'dftspinpol': 'F',
            'fullmatch': 'T',
            'l_nonsphdc': 'T',
            'l_correctetot': 'T',
            'l_forcehiaiteration': 'F',
        },
        'order': [],
        'several': set(),
        'simple': set(),
        'text': set(),
        'complex': set()
    }

    res, path = get_tag_info(schema_dict, 'ldaHIA', not_contains='atom')

    assert res == EXPECTED_RESULT
    assert path == '/fleurInput/calculationSetup/ldaHIA'

    with pytest.raises(ValueError, match='The tag ldaHIA has no possible paths with the current specification.'):
        res = get_tag_info(schema_dict, 'ldaHIA', not_contains='atom', contains='species')


def test_read_constants():
    """
    Test of the read_constants function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import read_constants

    VALID_INP_CONSTANTS_PATH = os.path.join(FILE_PATH, 'files/fleur/inp_with_constants.xml')
    INVALID_INP_CONSTANTS_PATH = os.path.join(FILE_PATH, 'files/fleur/inp_invalid_constants.xml')
    VALID_OUT_CONSTANTS_PATH = os.path.join(FILE_PATH, 'files/fleur/out_with_constants.xml')

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(VALID_INP_CONSTANTS_PATH, parser)
    outxmltree = etree.parse(VALID_OUT_CONSTANTS_PATH, parser)
    invalidxmltree = etree.parse(INVALID_INP_CONSTANTS_PATH, parser)

    root1 = xmltree.getroot()
    root2 = outxmltree.getroot()
    root3 = invalidxmltree.getroot()

    expected_constants = {
        'A': -3.14,
        'Ang': 1.889726124772898,
        'Bohr': 1.0,
        'Deg': 0.017453292519943295,
        'Pi': 3.141592653589793,
        'nm': 18.89726124772898,
        'notPi': 3.0,
        'pm': 0.01889726124772898
    }
    result = read_constants(root1, schema_dict_34)
    assert result == expected_constants

    result = read_constants(root2, outschema_dict_34)
    assert result == expected_constants

    with pytest.raises(KeyError, match='Ambiguous definition of constant Pi'):
        result = read_constants(root3, schema_dict_34)


def test_evaluate_attribute(caplog):
    """
    Test of the evaluate_attribute function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import evaluate_attribute

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    outxmltree = etree.parse(TEST_OUTXML_PATH2, parser)

    outroot = outxmltree.getroot()
    root = xmltree.getroot()

    assert evaluate_attribute(root, schema_dict, 'jspins', FLEUR_DEFINED_CONSTANTS) == 2
    assert evaluate_attribute(root, schema_dict, 'l_noco', FLEUR_DEFINED_CONSTANTS)
    assert evaluate_attribute(root, schema_dict, 'mode', FLEUR_DEFINED_CONSTANTS) == 'hist'
    assert pytest.approx(evaluate_attribute(root, schema_dict, 'radius', FLEUR_DEFINED_CONSTANTS,
                                            contains='species')) == [2.2, 2.2]

    with pytest.raises(ValueError, match='The attrib beta has multiple possible paths with the current specification.'):
        evaluate_attribute(root, schema_dict, 'beta', FLEUR_DEFINED_CONSTANTS, exclude=['unique'])

    assert pytest.approx(
        evaluate_attribute(root,
                           schema_dict,
                           'beta',
                           FLEUR_DEFINED_CONSTANTS,
                           exclude=['unique'],
                           contains='nocoParams',
                           not_contains='species')) == [np.pi / 2.0, np.pi / 2.0]

    assert pytest.approx(
        evaluate_attribute(root,
                           schema_dict,
                           'beta',
                           FLEUR_DEFINED_CONSTANTS,
                           tag_name='nocoParams',
                           not_contains='species')) == [np.pi / 2.0, np.pi / 2.0]

    with pytest.raises(ValueError, match='No attribute TEST found at tag nocoParams'):
        evaluate_attribute(
            root,
            schema_dict,
            'TEST',
            FLEUR_DEFINED_CONSTANTS,
            tag_name='nocoParams',
            not_contains='species',
        )

    assert pytest.approx(
        evaluate_attribute(outroot,
                           outschema_dict_34,
                           'beta',
                           FLEUR_DEFINED_CONSTANTS,
                           exclude=['unique'],
                           contains='nocoParams',
                           not_contains='species')) == [np.pi / 2.0, np.pi / 2.0]

    iteration = outroot.xpath('//iteration')[0]

    assert evaluate_attribute(iteration,
                              outschema_dict_34,
                              'units',
                              FLEUR_DEFINED_CONSTANTS,
                              tag_name='Forcetheorem_SSDISP') == 'Htr'

    with pytest.raises(ValueError, match='No attribute TEST found at tag Forcetheorem_SSDISP'):
        evaluate_attribute(iteration,
                           outschema_dict_34,
                           'TEST',
                           FLEUR_DEFINED_CONSTANTS,
                           tag_name='Forcetheorem_SSDISP')

    with pytest.raises(ValueError,
                       match='The attrib spinf has multiple possible paths with the current specification.'):
        evaluate_attribute(root, schema_dict, 'spinf', FLEUR_DEFINED_CONSTANTS)

    with pytest.raises(ValueError, match='No values found for attribute radius'):
        evaluate_attribute(root, schema_dict, 'radius', FLEUR_DEFINED_CONSTANTS, not_contains='species')

    with caplog.at_level(logging.WARNING):
        assert evaluate_attribute(
            root, schema_dict, 'radius', FLEUR_DEFINED_CONSTANTS, not_contains='species', logger=LOGGER) is None
    assert 'No values found for attribute radius' in caplog.text


def test_evaluate_text(caplog):
    """
    Test of the evaluate_text function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import evaluate_text

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    outxmltree = etree.parse(TEST_OUTXML_PATH2, parser)

    outroot = outxmltree.getroot()
    root = xmltree.getroot()

    assert pytest.approx(evaluate_text(root, schema_dict, 'qss', FLEUR_DEFINED_CONSTANTS)) == [0.0, 0.0, 0.0]
    kpoints = evaluate_text(root, schema_dict, 'kPoint', FLEUR_DEFINED_CONSTANTS)
    expected = [[-0.25, 0.25, 0.0], [0.25, 0.25, 0.0]]
    for val, result in zip(kpoints, expected):
        assert pytest.approx(val) == result

    positions = evaluate_text(root, schema_dict, 'filmPos', FLEUR_DEFINED_CONSTANTS)
    positions_out = evaluate_text(outroot, outschema_dict_34, 'filmPos', FLEUR_DEFINED_CONSTANTS)

    expected = [[0.0, 0.0, -0.9964250044], [0.5, 0.5, 0.9964250044]]
    for val, result in zip(positions, expected):
        assert pytest.approx(val) == result
    for val, result in zip(positions_out, expected):
        assert pytest.approx(val) == result

    with pytest.raises(ValueError, match='The tag row-1 has multiple possible paths with the current specification.'):
        evaluate_text(root, schema_dict, 'row-1', FLEUR_DEFINED_CONSTANTS)
    #Test the correct passing of contains parameters

    sym_row1 = evaluate_text(root, schema_dict, 'row-1', FLEUR_DEFINED_CONSTANTS, contains='symOp')
    expected = [[1.0, 0.0, 0.0, 0.0], [1.0, 0.0, 0.0, 0.0]]
    for val, result in zip(sym_row1, expected):
        assert pytest.approx(val) == result
    assert pytest.approx(
        evaluate_text(root,
                      schema_dict,
                      'row-1',
                      FLEUR_DEFINED_CONSTANTS,
                      not_contains='symOp',
                      contains='filmLattice/bravaisMatrix')) == [5.3011797029, 0.0, 0.0]

    with pytest.raises(ValueError, match='The tag TEST has no possible paths with the current specification.'):
        evaluate_text(root, schema_dict, 'TEST', FLEUR_DEFINED_CONSTANTS)

    with pytest.raises(ValueError, match='No text found for tag magnetism'):
        evaluate_text(root, schema_dict, 'magnetism', FLEUR_DEFINED_CONSTANTS, not_contains='species')

    with caplog.at_level(logging.WARNING):
        assert evaluate_text(
            root, schema_dict, 'magnetism', FLEUR_DEFINED_CONSTANTS, not_contains='species', logger=LOGGER) is None
    assert 'No text found for tag magnetism' in caplog.text


def test_evaluate_tag(caplog):
    """
    Test of the evaluate_tag function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import evaluate_tag

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    outxmltree = etree.parse(TEST_OUTXML_PATH2, parser)

    outroot = outxmltree.getroot()
    root = xmltree.getroot()

    expected = {
        'itmax': 1,
        'maxIterBroyd': 99,
        'imix': 'Anderson',
        'alpha': 0.05,
        'precondParam': 0.0,
        'spinf': 2.0,
        'minDistance': 1e-05,
        'maxTimeToStartIter': None
    }
    scfloop = evaluate_tag(root, schema_dict, 'scfLoop', FLEUR_DEFINED_CONSTANTS)
    assert scfloop == expected

    with pytest.raises(ValueError, match='The tag TEST has no possible paths with the current specification.'):
        evaluate_tag(root, schema_dict, 'TEST', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError,
                       match='The tag mtSphere has multiple possible paths with the current specification.'):
        evaluate_tag(root, schema_dict, 'mtSphere', FLEUR_DEFINED_CONSTANTS)

    with pytest.raises(ValueError, match='Failed to evaluate attributes from tag qss'):
        evaluate_tag(root, schema_dict, 'qss', FLEUR_DEFINED_CONSTANTS)

    with caplog.at_level(logging.WARNING):
        assert evaluate_tag(root, schema_dict, 'qss', FLEUR_DEFINED_CONSTANTS, logger=LOGGER) == {}
    assert 'Failed to evaluate attributes from tag qss' in caplog.text

    expected = {'radius': [2.2, 2.2], 'gridPoints': [787, 787], 'logIncrement': [0.016, 0.017]}
    mtRadii = evaluate_tag(root, schema_dict, 'mtSphere', FLEUR_DEFINED_CONSTANTS, contains='species')
    assert mtRadii == expected

    mtRadii = evaluate_tag(outroot, outschema_dict_34, 'mtSphere', FLEUR_DEFINED_CONSTANTS, contains='species')
    assert mtRadii == expected

    expected = {
        'l_constrained': None,
        'l_mtNocoPot': None,
        'l_relaxSQA': None,
        'l_magn': None,
        'M': None,
        'alpha': [0.0, 0.0],
        'beta': [1.570796326, 1.570796326],
        'b_cons_x': None,
        'b_cons_y': None
    }
    nocoParams = evaluate_tag(root, schema_dict, 'nocoParams', FLEUR_DEFINED_CONSTANTS, contains='atomGroup')
    assert nocoParams == expected

    expected = {
        'alpha': [0.0, 0.0],
        'beta': [1.570796326, 1.570796326],
    }
    nocoParams = evaluate_tag(root,
                              schema_dict,
                              'nocoParams',
                              FLEUR_DEFINED_CONSTANTS,
                              contains='atomGroup',
                              only_required=True)
    assert nocoParams == expected

    expected = {
        'beta': [1.570796326, 1.570796326],
    }
    nocoParams = evaluate_tag(root,
                              schema_dict,
                              'nocoParams',
                              FLEUR_DEFINED_CONSTANTS,
                              contains='atomGroup',
                              only_required=True,
                              ignore=['alpha'])
    assert nocoParams == expected


def test_single_value_tag(caplog):
    """
    Test of the evaluate_single_value_tag function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import evaluate_single_value_tag, get_tag_xpath
    from masci_tools.util.xml.common_functions import eval_xpath

    schema_dict = outschema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_OUTXML_PATH, parser)
    root = xmltree.getroot()

    iteration_xpath = get_tag_xpath(schema_dict, 'iteration')
    iteration = eval_xpath(root, iteration_xpath, list_return=True)[0]

    expected = {'comment': None, 'units': 'Htr', 'value': -4204.714048254}
    totalEnergy = evaluate_single_value_tag(iteration, schema_dict, 'totalEnergy', FLEUR_DEFINED_CONSTANTS)
    assert totalEnergy == expected

    expected = {'value': -4204.714048254}
    totalEnergy = evaluate_single_value_tag(iteration,
                                            schema_dict,
                                            'totalEnergy',
                                            FLEUR_DEFINED_CONSTANTS,
                                            only_required=True)
    assert totalEnergy == expected

    with pytest.raises(ValueError, match='The tag total_energy has no possible paths with the current specification.'):
        evaluate_single_value_tag(root, schema_dict, 'total_energy', FLEUR_DEFINED_CONSTANTS)
    with pytest.raises(ValueError,
                       match='The tag totalCharge has multiple possible paths with the current specification.'):
        evaluate_single_value_tag(root, schema_dict, 'totalCharge', FLEUR_DEFINED_CONSTANTS)

    expected = {'units': None, 'value': 63.9999999893}
    with pytest.raises(ValueError,
                       match="Failed to evaluate singleValue from tag totalCharge: Has no 'units' attribute"):
        evaluate_single_value_tag(iteration,
                                  schema_dict,
                                  'totalCharge',
                                  FLEUR_DEFINED_CONSTANTS,
                                  contains='allElectronCharges',
                                  not_contains='fixed')

    with caplog.at_level(logging.WARNING):
        totalCharge = evaluate_single_value_tag(iteration,
                                                schema_dict,
                                                'totalCharge',
                                                FLEUR_DEFINED_CONSTANTS,
                                                contains='allElectronCharges',
                                                not_contains='fixed',
                                                logger=LOGGER)
    assert totalCharge == expected
    assert "Failed to evaluate singleValue from tag totalCharge: Has no 'units' attribute" in caplog.text

    expected = {'value': 63.9999999893}
    totalCharge = evaluate_single_value_tag(iteration,
                                            schema_dict,
                                            'totalCharge',
                                            FLEUR_DEFINED_CONSTANTS,
                                            contains='allElectronCharges',
                                            not_contains='fixed',
                                            ignore=['units'])
    assert totalCharge == expected


def test_evaluate_parent_tag():
    """
    Test of the evaluate_parent_tag function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import evaluate_parent_tag

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_OUTXML_PATH, parser)
    root = xmltree.getroot()

    expected = {
        'atomicNumber': [31, 31, 33, 33],
        'element': ['Ga', 'Ga', 'As', 'As'],
        'name': ['Ga-1', 'Ga-1', 'As-2', 'As-2']
    }
    ldaU_species = evaluate_parent_tag(root, outschema_dict_34, 'ldaU', FLEUR_DEFINED_CONSTANTS, contains='species')
    pprint(ldaU_species)
    assert ldaU_species == expected

    expected = {'atomicNumber': [31, 31, 33, 33], 'name': ['Ga-1', 'Ga-1', 'As-2', 'As-2']}
    ldaU_species = evaluate_parent_tag(root,
                                       outschema_dict_34,
                                       'ldaU',
                                       FLEUR_DEFINED_CONSTANTS,
                                       contains='species',
                                       only_required=True)
    pprint(ldaU_species)
    assert ldaU_species == expected

    expected = {'name': ['Ga-1', 'Ga-1', 'As-2', 'As-2']}
    ldaU_species = evaluate_parent_tag(root,
                                       outschema_dict_34,
                                       'ldaU',
                                       FLEUR_DEFINED_CONSTANTS,
                                       contains='species',
                                       only_required=True,
                                       ignore=['atomicNumber'])
    pprint(ldaU_species)
    assert ldaU_species == expected


def test_tag_exists():
    """
    Test of the tag_exists function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import tag_exists

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    outxmltree = etree.parse(TEST_OUTXML_PATH2, parser)

    outroot = outxmltree.getroot()
    root = xmltree.getroot()

    assert tag_exists(root, schema_dict, 'filmPos')
    assert tag_exists(root, schema_dict, 'calculationSetup')

    assert not tag_exists(root, schema_dict, 'ldaU', contains='species')
    assert tag_exists(root, schema_dict, 'ldaU', not_contains='atom')

    assert not tag_exists(outroot, outschema_dict_34, 'ldaU', contains='species')
    assert tag_exists(outroot, outschema_dict_34, 'ldaU', not_contains='atom')

    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        tag_exists(root, schema_dict, 'ldaU')
    with pytest.raises(ValueError, match='The tag ldaU has no possible paths with the current specification.'):
        tag_exists(root, schema_dict, 'ldaU', contains='group')


def test_attrib_exists():
    """
    Test of the tag_exists function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import attrib_exists

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    outxmltree = etree.parse(TEST_OUTXML_PATH2, parser)

    outroot = outxmltree.getroot()
    root = xmltree.getroot()

    assert attrib_exists(root, schema_dict, 'itmax')
    assert attrib_exists(root, schema_dict, 'jspins')

    assert not attrib_exists(root, schema_dict, 'radius', contains='atomGroup')
    assert not attrib_exists(outroot, outschema_dict_34, 'radius', contains='atomGroup')

    with pytest.raises(ValueError,
                       match='The attrib spinf has multiple possible paths with the current specification.'):
        attrib_exists(root, schema_dict, 'spinf')
    with pytest.raises(ValueError, match='The attrib spinf has no possible paths with the current specification.'):
        attrib_exists(root, schema_dict, 'spinf', contains='group')


def test_get_number_of_nodes():
    """
    Test of the get_number_of_nodes function
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import get_number_of_nodes

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    xmltree = etree.parse(TEST_INPXML_PATH, parser)
    outxmltree = etree.parse(TEST_OUTXML_PATH2, parser)

    outroot = outxmltree.getroot()
    root = xmltree.getroot()

    assert get_number_of_nodes(root, schema_dict, 'filmPos') == 2
    assert get_number_of_nodes(root, schema_dict, 'calculationSetup') == 1
    assert get_number_of_nodes(root, schema_dict, 'ldaU', contains='species') == 0
    assert get_number_of_nodes(root, schema_dict, 'ldaU', not_contains='atom') == 1

    assert get_number_of_nodes(outroot, outschema_dict_34, 'filmPos') == 2
    assert get_number_of_nodes(outroot, outschema_dict_34, 'calculationSetup') == 1

    with pytest.raises(ValueError, match='The tag ldaU has multiple possible paths with the current specification.'):
        get_number_of_nodes(root, schema_dict, 'ldaU')
    with pytest.raises(ValueError, match='The tag ldaU has no possible paths with the current specification.'):
        get_number_of_nodes(root, schema_dict, 'ldaU', contains='group')


def test_schema_dict_util_abs_to_rel_path():
    """
    Test of the absolute to relative xpath conversion in schema_dict_util functions
    """
    from lxml import etree
    from masci_tools.util.schema_dict_util import eval_simple_xpath, get_number_of_nodes, tag_exists, \
                                                  evaluate_attribute, evaluate_tag, evaluate_parent_tag, \
                                                  evaluate_text

    schema_dict = schema_dict_34

    parser = etree.XMLParser(attribute_defaults=True, recover=False, encoding='utf-8')
    root = etree.parse(TEST_INPXML_PATH, parser).getroot()

    species = eval_simple_xpath(root, schema_dict, 'species')

    assert tag_exists(species[0], schema_dict, 'lo')
    assert tag_exists(species[1], schema_dict, 'lo')

    assert get_number_of_nodes(species[0], schema_dict, 'lo') == 2
    assert get_number_of_nodes(species[1], schema_dict, 'lo') == 1

    assert evaluate_attribute(species[0], schema_dict, 'name', constants=FLEUR_DEFINED_CONSTANTS) == 'Fe-1'

    assert evaluate_text(species[0], schema_dict, 'coreConfig', constants=FLEUR_DEFINED_CONSTANTS) == ['[Ne]']

    assert evaluate_tag(species[1], schema_dict, 'lo', constants=FLEUR_DEFINED_CONSTANTS) == {
        'eDeriv': 0,
        'l': 1,
        'n': 5,
        'type': 'SCLO'
    }
    assert evaluate_parent_tag(species[1], schema_dict, 'lo', constants=FLEUR_DEFINED_CONSTANTS) == {
        'atomicNumber': 78,
        'element': 'Pt',
        'name': 'Pt-1'
    }
